from configparser import ConfigParser

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.log import Logger
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.system.auxiliary.water_heating.water_heater import WaterHeating
from simses.technology.hydrogen.electrolyzer.degradation.no_degradation import NoDegradationModel
from simses.technology.hydrogen.electrolyzer.degradation.pem_analytic import \
    PemElMultiDimAnalyticDegradationModel
from simses.technology.hydrogen.electrolyzer.degradation.pem_analytic_ptl_coating import \
    PemElMultiDimAnalyticDegradationModelPtlCoating
from simses.technology.hydrogen.electrolyzer.pressure.abstract_model import PressureModel
from simses.technology.hydrogen.electrolyzer.pressure.alkaline import AlkalinePressureModel
from simses.technology.hydrogen.electrolyzer.pressure.no_pressure_model import \
    NoPressureModel
from simses.technology.hydrogen.electrolyzer.pressure.var_cathode import \
    VarCathodePressureModel
from simses.technology.hydrogen.electrolyzer.stack.alkaline.electrolyzer_model import \
    AlkalineElectrolyzer
from simses.technology.hydrogen.electrolyzer.stack.no_electrolyzer import NoElectrolyzer
from simses.technology.hydrogen.electrolyzer.stack.pem import PemElectrolyzer
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.electrolyzer_model import \
    PemElectrolyzerMultiDimAnalytic
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.electrolyzer_model_ptl_coating import \
    PemElectrolyzerMultiDimAnalyticPtlCoating
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel
from simses.technology.hydrogen.electrolyzer.thermal.abstract_model import ThermalModel
from simses.technology.hydrogen.electrolyzer.thermal.alkaline import \
    SimpleThermalModelAlkaline
from simses.technology.hydrogen.electrolyzer.thermal.no_thermal_model import NoThermalModel
from simses.technology.hydrogen.electrolyzer.thermal.pem import SimpleThermalModel


class ElectrolyzerFactory:

    def __init__(self, config: ConfigParser):
        self.__log: Logger = Logger(type(self).__name__)
        self.__config_general: GeneralSimulationConfig = GeneralSimulationConfig(config)
        self.__config_electrolyzer: ElectrolyzerConfig = ElectrolyzerConfig(config)
        self.__config_electrolyzer_data: ElectrolyzerDataConfig = ElectrolyzerDataConfig()

    def create_state(self, system_id: int, storage_id: int, temperature: float,
                     electrolyzer: ElectrolyzerStackModel, pressure_model: PressureModel) -> ElectrolyzerState:
        state: ElectrolyzerState = ElectrolyzerState(system_id, storage_id)
        state.time = self.__config_general.start
        state.power = 0  # W
        state.temperature = temperature - 273.15  # K -> °C
        state.power_loss = 0  # W
        state.hydrogen_production = 0  # mol/s
        state.hydrogen_outflow = 0  # mol/s
        state.oxygen_production = 0  # mol/s
        state.fulfillment = 1.0  # p.u.
        state.soh = 1.0  # p.u.
        state.convection_heat = 0  # W
        state.resistance_increase_cyclic = 0  # p.u.
        state.resistance_increase_calendar = 0  # p.u.
        state.resistance_increase = 0  # p.u.
        state.exchange_current_decrease_cyclic = 1  # p.u.
        state.exchange_current_decrease_calendar = 1  # p.u.
        state.exchange_current_decrease = 1  # p.u.
        state.pressure_anode = pressure_model.get_pressure_anode()  # barg
        state.pressure_cathode = pressure_model.get_pressure_cathode()  # barg
        state.sat_pressure_h2o = 0.03189409713071401  # bar  h2o saturation pressure at 25°C
        state.part_pressure_h2 = (1.0 + state.pressure_cathode) - 0.03189409713071401  # bar partial pressure h2 at 25°C and cathode pressure
        state.part_pressure_o2 = (1.0 + state.pressure_anode) - 0.03189409713071401  # bar partial pressure o2 at 25°C and anode pressure
        state.water_use = 0  # mol/s#
        state.water_outflow_cathode = 0  # mol/s
        state.water_outflow_anode = 0  # mol/s
        state.water_flow = 0  # mol/s
        state.power_water_heating = 0  # W
        state.power_pump = 0  # W
        state.power_gas_drying = 0  # W
        state.power_compressor = 0  # W
        state.total_hydrogen_production = 0  # kg
        state.relative_time = 0  # start
        state.voltage = 1  # stays at 1 so that electrolyzer and fuel cell always see the power indepentently from the voltage a timestep before
        electrolyzer.calculate(state.power, state)
        state.voltage = electrolyzer.get_voltage()
        state.current = electrolyzer.get_current()
        state.current_density = electrolyzer.get_current_density()
        state.max_charge_power = electrolyzer.get_nominal_stack_power()
        state.max_discharge_power = electrolyzer.get_nominal_stack_power()
        return state

    def create_electrolyzer_stack(self, electrolyzer: str, electrolyzer_maximal_power: float) -> ElectrolyzerStackModel:
        if electrolyzer == NoElectrolyzer.__name__:
            self.__log.debug('Creating electrolyzer as ' + electrolyzer)
            return NoElectrolyzer()
        elif electrolyzer == PemElectrolyzer.__name__:
            self.__log.debug('Creating electrolyzer as ' + electrolyzer)
            return PemElectrolyzer(electrolyzer_maximal_power, self.__config_electrolyzer_data)
        elif electrolyzer == PemElectrolyzerMultiDimAnalytic.__name__:
            self.__log.debug('Creating electrolyzer as ' + electrolyzer)
            return PemElectrolyzerMultiDimAnalytic(electrolyzer_maximal_power, self.__config_electrolyzer_data)
        elif electrolyzer == PemElectrolyzerMultiDimAnalyticPtlCoating.__name__:
            self.__log.debug('Creating electrolyzer as ' + electrolyzer)
            return PemElectrolyzerMultiDimAnalyticPtlCoating(electrolyzer_maximal_power, self.__config_electrolyzer_data)
        elif electrolyzer == AlkalineElectrolyzer.__name__:
            self.__log.debug('Creating electrolyzer as ' + electrolyzer)
            return AlkalineElectrolyzer(electrolyzer_maximal_power, self.__config_electrolyzer_data)
        else:
            options: [str] = list()
            options.append(PemElectrolyzer.__name__)
            raise Exception('Specified electrolyzer ' + electrolyzer + ' is unknown. '
                            'Following options are available: ' + str(options))

    def create_thermal_model(self, electrolyzer: ElectrolyzerStackModel, water_heating: WaterHeating,
                             pump: Pump, temperature: float) -> ThermalModel:
        if type(electrolyzer).__name__ in [PemElectrolyzerMultiDimAnalytic.__name__,
                                     PemElectrolyzerMultiDimAnalyticPtlCoating.__name__]:
            self.__log.debug('Creating electrolyzer thermal model as ' + SimpleThermalModel.__name__)
            return SimpleThermalModel(electrolyzer, water_heating, pump, temperature, self.__config_electrolyzer)
        elif type(electrolyzer).__name__ == AlkalineElectrolyzer.__name__:
           self.__log.debug('Creating electrolyzer thermal model as ' + SimpleThermalModelAlkaline.__name__)
           return SimpleThermalModelAlkaline(electrolyzer, water_heating, pump, temperature, self.__config_electrolyzer)
        else:
            self.__log.debug('Creating electrolyzer thermal model as ' + NoThermalModel.__name__)
            return NoThermalModel()

    def create_pressure_model(self, electrolyzer: ElectrolyzerStackModel) -> PressureModel:
        if type(electrolyzer).__name__ in [PemElectrolyzerMultiDimAnalytic.__name__,
                                     PemElectrolyzerMultiDimAnalyticPtlCoating.__name__] :
            self.__log.debug('Creating electrolyzer pressure model as ' + VarCathodePressureModel.__name__)
            return VarCathodePressureModel(electrolyzer, self.__config_electrolyzer)
        elif type(electrolyzer).__name__ == AlkalineElectrolyzer.__name__:
            self.__log.debug('Creating electrolyzer pressure model as ' + AlkalineElectrolyzer.__name__)
            return AlkalinePressureModel(self.__config_electrolyzer)
        else:
            self.__log.debug('Creating electrolyzer pressure model as ' + NoPressureModel.__name__)
            return NoPressureModel()

    def create_degradation_model(self, electrolyzer: ElectrolyzerStackModel):
        electrolyzer_name: str = type(electrolyzer).__name__
        if electrolyzer_name == PemElectrolyzerMultiDimAnalytic.__name__:
            self.__log.debug('Creating electrolyzer degradation model for ' + electrolyzer_name)
            return PemElMultiDimAnalyticDegradationModel(electrolyzer, self.__config_electrolyzer, self.__config_general)
        elif electrolyzer_name == PemElectrolyzerMultiDimAnalyticPtlCoating.__name__:
            self.__log.debug('Creating electrolyzer degradation model for ' + electrolyzer_name)
            return PemElMultiDimAnalyticDegradationModelPtlCoating(electrolyzer, self.__config_electrolyzer,
                                                                   self.__config_general)
        else:
            return NoDegradationModel()

    def close(self):
        self.__log.close()



from configparser import ConfigParser
from simses.commons.config.analysis.analysis_config import AnalysisConfig


class MarketProfileConfig(AnalysisConfig):

    """
    Market profile configs
    """

    SECTION: str = 'MARKET_PROFILE'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)
        # self.__log: Logger = Logger(type(self).__name__)

    @property
    def market_profile_dir(self) -> str:
        """Returns directory of market profiles from __analysis_config file_name"""
        return self.get_data_path(self.get_property(self.SECTION, 'MARKET_PROFILE_DIR'))

    @property
    def fcr_price_file(self) -> str:
        """Returns soc profile file_name name from __analysis_config file_name"""
        return self.market_profile_dir + self.get_property(self.SECTION, 'FCR_PRICE_PROFILE')

    @property
    def intraday_price_file(self) -> str:
        """ Return PV generation profile file_name name from __analysis_config file_name"""
        return self.market_profile_dir + self.get_property(self.SECTION, 'INTRADAY_PRICE_PROFILE')




from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties


class DefaultElectricalCellProperties(ElectricalCellProperties):

    __CELL_VOLTAGE = 3.5  # V
    __CELL_CAPACITY = 2.5  # Ah
    __MAX_VOLTAGE: float = 4.0  # V
    __MIN_VOLTAGE: float = 3.0  # V
    __MAX_CHARGE_RATE: float = 2.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 2.0  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
    __COULOMB_EFFICIENCY: float = 1.0  # p.u.

    def __init__(self):
        super(DefaultElectricalCellProperties, self).__init__(self.__CELL_VOLTAGE, self.__CELL_CAPACITY,
                                                              self.__MIN_VOLTAGE, self.__MAX_VOLTAGE,
                                                              self.__MAX_CHARGE_RATE, self.__MAX_DISCHARGE_RATE,
                                                              self.__SELF_DISCHARGE_RATE, self.__COULOMB_EFFICIENCY)



from configparser import ConfigParser

from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.simulation_config import clean_split
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.config.data.energy_management import EnergyManagementDataConfig
from simses.commons.log import Logger
from simses.commons.profile.power.alternating import AlternatePowerProfile
from simses.commons.profile.power.constant import ConstantPowerProfile
from simses.commons.profile.power.generation import GenerationProfile
from simses.commons.profile.power.load import LoadProfile
from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.profile.power.v2g_profile import V2GProfile
from simses.commons.profile.technical.binary import BinaryProfile
from simses.commons.profile.power.random import RandomPowerProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.logic.energy_management.strategy.basic.ev_charger_with_buffer import EvChargerWithBuffer
from simses.logic.energy_management.strategy.basic.electric_vehicle import ElectricVehicle
from simses.logic.energy_management.strategy.basic.frequency_containment_reserve import FrequencyContainmentReserve
from simses.logic.energy_management.strategy.basic.intraday_market_recharge import IntradayMarketRecharge
from simses.logic.energy_management.strategy.basic.peak_shaving_perfect_foresight import PeakShavingPerfectForesight
from simses.logic.energy_management.strategy.basic.peak_shaving_simple import SimplePeakShaving
from simses.logic.energy_management.strategy.basic.power_follower import PowerFollower
from simses.logic.energy_management.strategy.basic.residential_pv_feed_in_damp import ResidentialPvFeedInDamp
from simses.logic.energy_management.strategy.basic.residential_pv_greedy import ResidentialPvGreedy
from simses.logic.energy_management.strategy.basic.soc_follower import SocFollower
from simses.logic.energy_management.strategy.basic.electric_vehicle_soc import ElectricVehicleSOC
from simses.logic.energy_management.strategy.basic.electric_vehicle_v2g import ElectricVehicleV2G
from simses.logic.energy_management.strategy.basic.use_all_renewable_energy import UseAllRenewableEnergy
from simses.logic.energy_management.strategy.stacked.fcr_idm_recharge_stacked import FcrIdmRechargeStacked
from simses.logic.energy_management.strategy.stacked.parallel_multi_use_with_stealing import ParallelMultiUseWithStealing

class EnergyManagementFactory:
    """
    Energy Management Factory to create the operation strategy of the ESS.
    """

    def __init__(self, config: ConfigParser, path: str = None):
        self.__log: Logger = Logger(type(self).__name__)
        self.__config_general: GeneralSimulationConfig = GeneralSimulationConfig(config, path)
        self.__config_ems: EnergyManagementConfig = EnergyManagementConfig(config, path)
        self.__config_profile: ProfileConfig = ProfileConfig(config, path)
        self.__config_system: StorageSystemConfig = StorageSystemConfig(config, path)
        self.__config_battery: BatteryConfig = BatteryConfig(config, path)
        self.__config_ems_data: EnergyManagementDataConfig = EnergyManagementDataConfig()

    def create_operation_strategy(self):
        """
        Energy Management Factory to create the operation strategy of the ESS based on the __analysis_config file_name.
        """
        os = self.__config_ems.operation_strategy
        timestep = self.__config_general.timestep

        if os == FrequencyContainmentReserve.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return FrequencyContainmentReserve(self.__config_general, self.__config_ems, self.__config_profile)

        elif os == IntradayMarketRecharge.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return IntradayMarketRecharge(self.__config_general, self.__config_ems)

        elif os == SimplePeakShaving.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return SimplePeakShaving(self.load_profile(), self.__config_ems)

        elif os == PeakShavingPerfectForesight.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return PeakShavingPerfectForesight(self.__config_general, self.load_profile(), self.load_profile(),
                                               self.__config_ems, self.__config_system, self.__config_profile)

        elif os == PowerFollower.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return PowerFollower(self.load_profile())

        elif os == SocFollower.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return SocFollower(self.__config_general, self.__config_profile)

        elif os == FcrIdmRechargeStacked.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return FcrIdmRechargeStacked(self.__config_general, self.__config_ems, self.__config_profile)

        elif os == ResidentialPvGreedy.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return ResidentialPvGreedy(self.load_profile(), self.generation_profile())

        elif os == ResidentialPvFeedInDamp.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return ResidentialPvFeedInDamp(self.load_profile(), self.__config_general, self.__config_profile)

        elif os == UseAllRenewableEnergy.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return UseAllRenewableEnergy(self.generation_profile())

        elif os == EvChargerWithBuffer.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return EvChargerWithBuffer(self.load_profile(), self.__config_ems)

        elif os == ElectricVehicle.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return ElectricVehicle(self.__config_general, self.__config_profile, self.__config_ems, self.load_profile())

        elif os == ElectricVehicleSOC.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return ElectricVehicleSOC(self.__config_general, self.__config_profile, self.__config_ems)

        elif os == ElectricVehicleV2G.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return ElectricVehicleV2G(self.__config_general, self.__config_profile, self.__config_ems, self.load_profile(),
                                      self.v2g_profile())

        elif os == ParallelMultiUseWithStealing.__name__:
            self.__log.debug('Creating operation strategy as ' + os)
            return ParallelMultiUseWithStealing(self.__config_general, self.__config_profile, self.__config_ems, self.load_profile(),
                                   self.__config_battery, self.generation_profile(), self.__config_system)

        else:
            options: [str] = list()
            options.append(FrequencyContainmentReserve.__name__)
            options.append(IntradayMarketRecharge.__name__)
            options.append(SimplePeakShaving.__name__)
            options.append(PeakShavingPerfectForesight.__name__)
            options.append(PowerFollower.__name__)
            options.append(SocFollower.__name__)
            options.append(FcrIdmRechargeStacked.__name__)
            options.append(ResidentialPvGreedy.__name__)
            options.append(ResidentialPvFeedInDamp.__name__)
            options.append(EvChargerWithBuffer.__name__)
            options.append(ElectricVehicle.__name__)
            options.append(ElectricVehicleSOC.__name__)
            options.append(ElectricVehicleV2G.__name__)
            options.append(ParallelMultiUseWithStealing.__name__)
            raise Exception('Operation strategy ' + os + ' is unknown. '
                                                         'Following options are available: ' + str(options))

    def generation_profile(self) -> GenerationProfile:
        return GenerationProfile(self.__config_profile, self.__config_general)

    def binary_profile(self) -> BinaryProfile:
        return BinaryProfile(self.__config_general, self.__config_profile)

    def v2g_profile(self) -> PowerProfile:
        return V2GProfile(self.__config_profile, self.__config_general)

    def load_profile(self) -> PowerProfile:
        """
        Returns the load profile for the EnergyManagementSystem
        """
        power_profile = self.__config_profile.load_profile_file
        profile: [str] = clean_split(power_profile, ',')
        try:
            power: float = float(profile[1])
        except IndexError:
            power = None
        if RandomPowerProfile.__name__ in power_profile:
            try:
                power_offset: float = float(profile[2])
            except IndexError:
                power_offset: float = 0.0
            start: float = self.__config_general.start
            return RandomPowerProfile(start, max_power=1500.0 if power is None else power, power_offset=power_offset)
        elif ConstantPowerProfile.__name__ in power_profile:
            return ConstantPowerProfile(power=0.0 if power is None else power, scaling_factor=1)
        elif AlternatePowerProfile.__name__ in power_profile:
            try:
                power_off: float = float(profile[2])
                time_on: float = float(profile[3])
                time_off: float = float(profile[4])
            except IndexError:
                power_off = 0
                time_on = 6
                time_off = 6
            return AlternatePowerProfile(power_on=1500.0 if power is None else power, power_off=power_off,
                                         scaling_factor=-1, time_on=time_on, time_off=time_off)
        else:
            return LoadProfile(self.__config_profile, self.__config_general)

    def create_energy_management_state(self) -> EnergyManagementState:
        state: EnergyManagementState = EnergyManagementState()
        state.time = self.__config_general.start
        return state

    def close(self):
        self.__log.close()



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.technology.storage import StorageTechnologyState
from .cycle_detector import CycleDetector
from math import floor, log10


class HalfCycleDetector(CycleDetector):

    def __init__(self, start_soc: float, general_config: GeneralSimulationConfig):
        super().__init__()
        self.__total_simulation_time = general_config.end + (general_config.duration * (general_config.loop - 1))

        # cycle parameters
        self.__depth_of_cycle = 0
        self.__full_equivalent_cycles = 0
        self.__c_rate = 0
        self.__mean_soc: float = 0.0

        # cycle counting parameters
        self.__cycle_step = 1
        self.__last_soc = start_soc
        self.__start_soc = start_soc
        self.__flag_charging = True

        # set accuracy based on timestep width
        self.__delta_t = general_config.timestep
        c_rate_accuracy = 0.001  # used for rounding soc - detects charge/discharge defined here at given delta_t
        soc_delta = c_rate_accuracy * self.__delta_t / 3600
        if soc_delta == 0:
            self.__cycle_counting_accuracy = 0
        else:
            self.__cycle_counting_accuracy = -floor(log10(abs(soc_delta)))

    def cycle_detected(self, time: float, state: StorageTechnologyState) -> bool:
        soc = round(state.soc, self.__cycle_counting_accuracy)
        delta_t = time - state.time
        if (self.__flag_charging and soc > self.__last_soc) or (not self.__flag_charging and soc < self.__last_soc):
            cycle_detected = False
            self.__last_soc = soc
            self.__cycle_step += 1
        elif soc == self.__last_soc:
            cycle_detected = False
            # cycle_step is not increased for constant soc
        else:
            cycle_detected = True

        # Last simulation step reached
        if (time + delta_t) > self.__total_simulation_time:
            cycle_detected = True

        if cycle_detected:
            # determine depth of cycle, c_rate and average SOC for the completed half_cycle
            self.__depth_of_cycle = abs(self.__start_soc - self.__last_soc)
            self.__c_rate = abs(self.__start_soc - self.__last_soc) / (self.__cycle_step * delta_t)
            self.__mean_soc = (self.__start_soc + self.__last_soc) / 2.0
            # reset cycle parameters
            if soc < self.__last_soc: self.__flag_charging = False
            if soc > self.__last_soc: self.__flag_charging = True
            self.__start_soc = self.__last_soc
            self.__last_soc = soc
            self.__cycle_step = 1

        return cycle_detected

    def get_depth_of_cycle(self) -> float:
        return self.__depth_of_cycle

    def get_delta_full_equivalent_cycle(self) -> float:
        return self.__depth_of_cycle / 2

    def get_crate(self) -> float:
        return self.__c_rate

    def get_mean_soc(self) -> float:
        return self.__mean_soc

    def get_full_equivalent_cycle(self) -> float:
        # add new cycle
        self.__full_equivalent_cycles += self.__depth_of_cycle / 2
        return self.__full_equivalent_cycles

    def reset(self) -> None:
        pass



import numpy as np

from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class PeakShavingPerfectForesight(OperationStrategy):
    """
        Peak Shaving under the assumption of perfect foresight for the load profile in order to reduce calendar
        degradation.
        The BESS will only charge up to the energy that is needed for the next load peak, right before the load peak
        occurs.
    """
    def __init__(self, general_config: GeneralSimulationConfig, power_profile: PowerProfile, forecast_profile: PowerProfile,
                 energy_management_config: EnergyManagementConfig, system_config: StorageSystemConfig,
                 profile_config: ProfileConfig):
        super().__init__(OperationPriority.VERY_HIGH)

        # configurable
        self.__efficiency = 0.9
        self.__soc_margin = 0.1  # to prevent missing peaks due to non-accurate efficiency
        forecasting_time_hours = 72  # select forecast time for data horizon in the future that is considered
        forecasting_steps = int(round(forecasting_time_hours*60*60 / general_config.timestep))

        self.__load_profile = power_profile
        self.__delta_t = general_config.timestep
        self.__ps_limit: float = energy_management_config.max_power
        nameplate_power = 0
        for system in system_config.storage_systems_ac:
            nameplate_power += float(system[1])
        self.__peak_energy_forecaster = PeakEnergyForecaster(forecast_profile, forecasting_steps, self.__ps_limit, nameplate_power,
                                                             self.__efficiency, general_config, profile_config)
        self.__power = 0

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        # Only charge BESS to PeakShavingEnergy that is required during the next forecast cycle for reduced ageing.
        # If forecasting horizon is chosen too short, the BESS may not be able to charge up before the peak.
        self.__power = self.__load_profile.next(time) + power

        peak_shaving_energy_delta = -1 * self.__peak_energy_forecaster.next(time, self.__power) * 1/3600

        net_load = self.__power - self.__ps_limit
        if net_load >= 0:
            target_power = -1 * net_load
        else:
            if (system_state.soc - self.__soc_margin) * system_state.capacity < peak_shaving_energy_delta:
                # adapt power for long simulation steps to prevent overshoot
                # otherwise: target_power = -1 * net_load
                missing_wh = peak_shaving_energy_delta - (system_state.soc - self.__soc_margin) * system_state.capacity
                missing_wh *= 1 / self.__efficiency
                target_power = missing_wh * 3600 / self.__delta_t
                if target_power > -1 * net_load:
                    target_power = -1 * net_load
            elif self.__soc_margin > system_state.soc:
                missing_wh = (self.__soc_margin - system_state.soc) * system_state.capacity
                target_power = missing_wh * 3600 / self.__delta_t
                if target_power > -1 * net_load:
                    target_power = -1 * net_load
            else:
                target_power = 0

        return target_power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.load_power = self.__power
        energy_management_state.peakshaving_limit = self.__ps_limit

    def clear(self) -> None:
        self.__power = 0.0

    def close(self) -> None:
        pass


class PeakEnergyForecaster:

    def __init__(self, profile: PowerProfile, forecasting_steps: int, ps_limit: float, p_max: float,
                 efficiency: float, general_config: GeneralSimulationConfig, profile_config: ProfileConfig):

        # open second load profile for forecasts
        # self.__load_profile: LoadProfile = LoadProfile(profile_config, general_config)
        self.__load_profile: PowerProfile = profile

        # needed for algorithm
        self.__efficiency = efficiency
        self.__delta_t = general_config.timestep
        self.__forecasting_steps = forecasting_steps
        self.__start_t = general_config.start
        self.__nameplate_power = p_max
        self.__ps_limit = ps_limit
        power_values = list()
        # for i in range(forecasting_steps):
        for i in range(1, forecasting_steps+1):
            power_values.append(self.__load_profile.next(self.__start_t+i*self.__delta_t))

        # calculate initial energy surplus for every future peak within the load forecast timeframe in Ws
        self.__energy_surpluses = [0]
        self.__time_peak_end = []  # time at which the respective peak ends
        self.__sign = np.sign(ps_limit - power_values[0])
        for i in range(len(power_values)):
            delta_p = self.limit(ps_limit - power_values[i], self.__nameplate_power)
            if delta_p > 0:
                delta_p *= self.__efficiency
            if np.sign(delta_p) <= self.__sign:
                self.__energy_surpluses[-1] += delta_p * self.__delta_t
            else:
                self.__energy_surpluses.append(self.__energy_surpluses[-1])
                self.__time_peak_end.append(self.__start_t + (i+1) * self.__delta_t)
                self.__energy_surpluses[-1] += delta_p * self.__delta_t
            self.__sign = np.sign(delta_p)

    @staticmethod
    def limit(delta: float, max_abs: float) -> float:
        if delta < - max_abs:
            return -max_abs
        elif delta > max_abs:
            return max_abs
        else:
            return delta

    def next(self, time, current_load) -> float:
        # check if a peak has passed
        if self.__time_peak_end:
            if self.__time_peak_end[0] <= time:
                self.__time_peak_end.pop(0)
                self.__energy_surpluses.pop(0)

        # subtract current peak energy from energy of future peaks
        delta_p_now = self.limit(self.__ps_limit - current_load, self.__nameplate_power)
        if delta_p_now > 0:
            delta_p_now *= self.__efficiency
        self.__energy_surpluses = [e - delta_p_now * self.__delta_t for e in self.__energy_surpluses]

        # look for new peaks
        delta_p_future = self.limit(
            self.__ps_limit - self.__load_profile.next(time + self.__forecasting_steps * self.__delta_t),
            self.__nameplate_power)
        if delta_p_future > 0:
            delta_p_future *= self.__efficiency
        if np.sign(delta_p_future) <= self.__sign:
            self.__energy_surpluses[-1] += delta_p_future * self.__delta_t
        else:
            self.__energy_surpluses.append(self.__energy_surpluses[-1])
            self.__time_peak_end.append(time + self.__forecasting_steps * self.__delta_t)
            self.__energy_surpluses[-1] += delta_p_future * self.__delta_t
        self.__sign = np.sign(delta_p_future)
        return min(self.__energy_surpluses)



from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.load.dc_load import DcLoad


class NoDcLoad(DcLoad):

    def __init__(self):
        super().__init__()

    def get_power(self) -> float:
        return 0.0

    def calculate_power(self, time: float) -> None:
        pass

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self) -> None:
        pass



import pandas as pd

from simses.commons.log import Logger
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.profile.power.generation import GenerationProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.stacked.stacked_operation_strategy import StackedOperationStrategy
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.basic.frequency_containment_reserve import FrequencyContainmentReserve
from simses.logic.energy_management.strategy.stacked.fcr_idm_recharge_stacked import FcrIdmRechargeStacked
from simses.logic.energy_management.strategy.basic.peak_shaving_simple import SimplePeakShaving
from simses.logic.energy_management.strategy.basic.peak_shaving_perfect_foresight import PeakShavingPerfectForesight
from simses.logic.energy_management.strategy.basic.power_follower import PowerFollower
from simses.logic.energy_management.strategy.basic.soc_follower import SocFollower
from simses.logic.energy_management.strategy.basic.intraday_market_recharge import IntradayMarketRecharge
from simses.logic.energy_management.strategy.basic.residential_pv_greedy import ResidentialPvGreedy
from simses.logic.energy_management.strategy.basic.residential_pv_feed_in_damp import ResidentialPvFeedInDamp
import numpy as np


class ParallelMultiUseWithStealing(StackedOperationStrategy):
    # 0. Section: Initialize variables and strategies
    def __init__(self, general_config: GeneralSimulationConfig, profile_config: ProfileConfig,
                 ems_config: EnergyManagementConfig, power_profile: PowerProfile, battery_config: BatteryConfig,
                 pv_generation_profile: GenerationProfile, system_config: StorageSystemConfig):

        self.__log: Logger = Logger(type(self).__name__)
        self.__strategies_names: [] = ems_config.multi_use_strategies
        self.__energy_allocations: [] = ems_config.energy_allocation
        self.__energy_allocations = [float(item) for item in self.__energy_allocations]
        self.__power_allocations: [] = ems_config.power_allocation
        self.__power_allocations = [float(item) for item in self.__power_allocations]
        self.__ranking: [] = ems_config.multi_use_rank
        self.__ranking = [int(item) for item in self.__ranking]
        self.__start_soc = battery_config.soc
        self.__ts = general_config.timestep
        self.__ac_storage_system_number = 0
        self.__storage_power_installed = \
            float(system_config.storage_systems_ac[self.__ac_storage_system_number][system_config.AC_SYSTEM_POWER])

        self.__strategies = []
        self.__system_id = 0
        self.__storage_id = 0
        self.__sub_system_states = []
        self.__btm_ftm = []  # BTM = 0, FTM = 1

        self.__a = 0
        self.__b = 0

        self.__total_power = 0
        self.__power_sub_system_states = []
        self.__power_requested = []
        self.__power_allocated = []
        self.__borrowed_capacity = []

        self.__previous_system_state_soc = self.__start_soc
        self.__system_dc_name = \
            system_config.storage_systems_dc[self.__ac_storage_system_number][system_config.DC_SYSTEM_STORAGE]
        self.__previous_system_state_capacity = \
            float(system_config.storage_technologies[self.__system_dc_name][system_config.STORAGE_CAPACITY])

        self.__delta_system_state_soc = 0
        self.__delta_sub_system_state_soc = 0
        self.__delta_system_state_capacity = 0
        self.__delta_sub_system_state_capacity = 0

        self.__counter = 0

        # Sort strategies by ranking:
        strategies_nparray = np.array(self.__strategies_names)
        self.__strategies_names = strategies_nparray[np.array(self.__ranking).argsort()].tolist()
        self.__power_allocations = np.array(self.__power_allocations)[np.array(self.__ranking).argsort()].tolist()
        self.__energy_allocations = np.array(self.__energy_allocations)[np.array(self.__ranking).argsort()].tolist()

        # Create sub system states for each strategy:
        for strategy in self.__strategies_names:
            if strategy == 'PowerFollower':
                self.__strategies.append(PowerFollower(power_profile))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
            elif strategy == 'SocFollower':
                self.__strategies.append(SocFollower(general_config, profile_config))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
            elif strategy == 'FrequencyContainmentReserve':
                self.__strategies.append(FrequencyContainmentReserve(general_config, ems_config, profile_config))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(1)
            elif strategy == 'FcrIdmRechargeStacked':
                self.__strategies.append(FcrIdmRechargeStacked(general_config, ems_config, profile_config))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(1)
            elif strategy == 'SimplePeakShaving':
                self.__strategies.append(SimplePeakShaving(power_profile, ems_config))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(0)
            elif strategy == 'PeakShavingPerfectForesight':
                self.__strategies.append(PeakShavingPerfectForesight(general_config, power_profile, power_profile,
                                                                   ems_config, system_config, profile_config))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(0)
            elif strategy == 'IntradayMarketRecharge':
                self.__strategies.append(IntradayMarketRecharge(general_config, ems_config))
                self.__sub_system_states.append(SystemState(self.__system_id, self.__storage_id))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(1)
            elif strategy == 'ResidentialPvGreedy':
                self.__strategies.append(ResidentialPvGreedy(power_profile, pv_generation_profile))
                self.__sub_system_states.append((SystemState(self.__system_id, self.__storage_id)))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(0)
            elif strategy == 'ResidentialPvFeedInDamp':
                self.__strategies.append(ResidentialPvFeedInDamp(power_profile, general_config, pv_generation_profile))
                self.__sub_system_states.append((SystemState(self.__system_id, self.__storage_id)))
                self.__sub_system_states[self.__storage_id].soc = self.__start_soc
                self.__btm_ftm.append(0)
            self.__storage_id += 1

        # Warning if energy or power allocations are not equal 1:
        if sum(self.__energy_allocations) != 1 or sum(self.__power_allocations) != 1:
            # Warning if energy allocation exceeds storage limits
            if sum(self.__energy_allocations) > 1:
                raise ValueError('Energy allocation exceeds storage limits. ENERGY_ALLOCATION should be 1.')
            elif sum(self.__energy_allocations) < 1:
                raise ValueError('Energy allocation is less then storage limits. ENERGY_ALLOCATION should be 1.')
            # Warning if power allocation exceeds power limits
            if sum(self.__power_allocations) > 1:
                raise ValueError('Power allocation exceeds storage limits. POWER_ALLOCATION should be 1.')
            elif sum(self.__power_allocations) < 1:
                raise ValueError('Power allocation is less then storage limits. POWER_ALLOCATION should be 1.')

        # TODO: Nochmal mit drüber sprechen, weil 2 mal sortiert wird
        temp_strategies_backup = self.__strategies.copy()
        super().__init__(OperationPriority.VERY_HIGH, self.__strategies)  # TODO: Ask Marc, if necessary
        # Reorder self.__strategies:
        self.__strategies = temp_strategies_backup
        self.__borrowed_capacity = np.zeros(len(self.__strategies)).tolist()

        self.__indices_btm = [i for i, x in enumerate(self.__btm_ftm) if x == 0]  # Get indices of BTM strategy
        self.__indices_ftm = [i for i, x in enumerate(self.__btm_ftm) if x == 1]  # Get indices of FTM strategy

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__counter = self.__counter + 1
        # if self.__counter == 859:
            # print('Halt')

        self.__total_power = 0
        self.__power_requested = []

        # 1. Section: Calculate Sub-System-SOC
        if system_state.ac_power != 0:
            self.__a = system_state.ac_power_delivered / system_state.ac_power  # Ratio A: Delivered power divided by requested power

        self.__delta_system_state_soc = system_state.soc - self.__previous_system_state_soc
        self.__delta_system_state_capacity = self.__delta_system_state_soc * self.__previous_system_state_capacity

        for subsystem_counter in range(len(self.__strategies)):
            # Update Sub-System-States: capacity, max charge and max discharge power
            self.__sub_system_states[subsystem_counter].capacity = self.__previous_system_state_capacity * self.__energy_allocations[subsystem_counter]
            self.__sub_system_states[subsystem_counter].max_charge_power = self.__storage_power_installed * self.__power_allocations[subsystem_counter]
            self.__sub_system_states[subsystem_counter].max_discharge_power = self.__storage_power_installed * self.__power_allocations[subsystem_counter]

            # Delivered AC power of subsystem corresponds to ratio of total AC power (ratio A)
            self.__sub_system_states[subsystem_counter].ac_power_delivered = self.__sub_system_states[subsystem_counter].ac_power * self.__a

            # Ratio B: Delivered power of subsystem divided by delivered power of total system
            if system_state.ac_power_delivered != 0:
                self.__b = self.__sub_system_states[subsystem_counter].ac_power_delivered / system_state.ac_power_delivered
            else:
                self.__b = 1/len(self.__strategies)  # Losses are half for each application

            # Calculation of Subsystem-SoC using Ratio B
            self.__delta_sub_system_state_capacity = self.__delta_system_state_capacity * self.__b

            self.__sub_system_states[subsystem_counter].soc = self.__sub_system_states[subsystem_counter].soc \
                                                               + self.__delta_sub_system_state_capacity \
                                                               / self.__sub_system_states[subsystem_counter].capacity

            # 1.1 Section: Correct Sub-System-SOC if it is not between 0 and 1
            # If SOC is smaller than 0 or bigger than 1:
            # Energy has been borrowed from other application of same pot (BTM or FTM)
            # Or more energy has been discharged than expected. Why?
            sub_system_states = self.__sub_system_states.copy()
            btm_ftm = self.__btm_ftm.copy()
            self.__sub_system_states = self.correct_subsystem_soc(sub_system_states, subsystem_counter, btm_ftm,
                                                                  self.__indices_btm, self.__indices_ftm, system_state)

            # 1.2 Section: Consider degradation
            # Capacity is lower -> Subsystem-Capacity is lower
            delta_capacity = self.__previous_system_state_capacity - system_state.capacity
            self.__sub_system_states[subsystem_counter].capacity = self.__sub_system_states[subsystem_counter].capacity \
                                                                    - delta_capacity * self.__energy_allocations[subsystem_counter]

            # Section 2: Next iteration begins here
            # Call next function of strategy to calculate requested power
            power = self.__strategies[subsystem_counter].next(time, self.__sub_system_states[subsystem_counter])
            self.__power_requested.append(power)
            self.__sub_system_states[subsystem_counter].ac_power = power
            self.__total_power += power

        # Section 3: Check here,if subsystem-soc allows the desired power:
        self.__borrowed_capacity = np.zeros(len(self.__strategies)).tolist()
        sub_system_states = self.__sub_system_states.copy()
        btm_ftm = self.__btm_ftm.copy()
        power_requested = self.__power_requested.copy()
        borrowed_capacity = self.__borrowed_capacity.copy()
        self.__power_requested, self.__borrowed_capacity = \
            self.check_subsystem_soc_allows_power(self.__indices_btm, self.__indices_ftm, btm_ftm,
                                                  power_requested, sub_system_states, borrowed_capacity)

        # Section 4:
        # After calculation of all applications: Check if power allocation does not surpass maximum power for btm & ftm:
        # Calculate power of each subsystem under consideration of ranking, BTM/FTM and maximum power
        power_allocated = self.calculate_power_considering_ranking(self.__power_requested,
                                                            self.__storage_power_installed,
                                                            self.__btm_ftm)
        # Correct AC power:
        counter_power_allocated = 0
        for i in range(len(self.__strategies)):
            self.__sub_system_states[i].ac_power = power_allocated[counter_power_allocated]
            counter_power_allocated += 1

        # Correct total_power:
        self.__total_power = sum(power_allocated)
        # Save system state SOC and capacity for the next time step:
        self.__previous_system_state_soc = system_state.soc
        self.__previous_system_state_capacity = system_state.capacity

        # # Save subsystem_SOC here to CSV:
        # soc1 = self.__sub_system_states[0].soc
        # soc2 = self.__sub_system_states[1].soc
        # #soc3 = self.__sub_system_states[2].soc
        # cap1 = self.__sub_system_states[0].capacity
        # cap2 = self.__sub_system_states[1].capacity
        # #cap3 = self.__sub_system_states[2].capacity
        # power_requested_copy1 = power_requested[0]
        # power_requested_copy2 = power_requested[1]
        # #power_requested_copy3 = power_requested[2]
        # power_allocated_copy1 = power_allocated[0]
        # power_allocated_copy2 = power_allocated[1]
        # #power_allocated_copy3 = power_allocated[2]
        #
        # df = pd.DataFrame([soc1, soc2, cap1, cap2,
        #                    power_requested_copy1, power_requested_copy2,
        #                    power_allocated_copy1, power_allocated_copy2]).transpose()
        #
        # # df = pd.DataFrame([soc1, soc2, soc3, cap1, cap2, cap3,
        # #                    power_requested_copy1, power_requested_copy2, power_requested_copy3,
        # #                    power_allocated_copy1, power_allocated_copy2, power_allocated_copy3]).transpose()
        # temp_name = 'multi_use_sub_soc.csv'
        #
        # try:
        #    f = open(temp_name)
        #    df.to_csv(temp_name, mode='a', index=False, header=False)
        #
        #    f.close()
        # except IOError:
        #    df.to_csv(temp_name, header=False)  # Create CSV

        return self.__total_power

    def correct_subsystem_soc(self, sub_system_states, subsystem_counter, btm_ftm, indices_btm, indices_ftm, system_state):
        # Correction of Sub-System-State-SoF if System-State-SoC is 1 or 0
        if system_state.soc == 1:  # Set all subsystem-socs to one to prevent increasing deviation over time
            sub_system_states[subsystem_counter].soc = 1
        elif system_state.soc == 0:  # Set all subsystem-socs to zero to prevent increasing deviation over tim
            sub_system_states[subsystem_counter].soc = 0

        # Check if it possible for all BTM or FTM applications to borrow energy
        soc_sum_btm = sum(sub_system_states[j].soc for j in indices_btm)
        soc_sum_ftm = sum(sub_system_states[j].soc for j in indices_ftm)
        if soc_sum_btm < 0:
            print('Warning! Sub-System-SoC is bigger/smaller then 1/0! Sub-System-SoC;', sub_system_states[subsystem_counter].soc)
            for btm_systems in indices_btm:
                sub_system_states[btm_systems].soc = 0

        elif soc_sum_btm > len(indices_btm):
            print('Warning! Sub-System-SoC is bigger/smaller then 1/0! Sub-System-SoC;', sub_system_states[subsystem_counter].soc)
            for btm_systems in indices_btm:
                sub_system_states[btm_systems].soc = 1

        elif soc_sum_ftm < 0:
            print('Warning! Sub-System-SoC is bigger/smaller then 1/0! Sub-System-SoC;', sub_system_states[subsystem_counter].soc)
            for ftm_systems in indices_ftm:
                sub_system_states[ftm_systems].soc = 0

        elif soc_sum_ftm > len(indices_ftm):
            print('Warning! Sub-System-SoC is bigger/smaller then 1/0! Sub-System-SoC;', sub_system_states[subsystem_counter].soc)
            for ftm_systems in indices_ftm:
                sub_system_states[ftm_systems].soc = 1

        # If it is possible
        else:
            if sub_system_states[subsystem_counter].soc < 0:
                temp_delta = sub_system_states[subsystem_counter].soc * sub_system_states[subsystem_counter].capacity
                sub_system_states[subsystem_counter].soc = 0  # Set SOC to 0

                if btm_ftm[subsystem_counter] == 0:  # BTM
                    for btm_appl in [x for x in indices_btm if x > subsystem_counter]:  # go through next applications to take from them
                        btm_available_capacity_current = np.round(sub_system_states[btm_appl].soc *
                                                                  sub_system_states[btm_appl].capacity, 6)

                        if btm_available_capacity_current > 0 and btm_available_capacity_current >= -temp_delta:
                            # Reduce next BTM application SOC by desired capacity
                            sub_system_states[btm_appl].soc += temp_delta / sub_system_states[btm_appl].capacity
                            temp_delta = 0  # because complete temp_delta is taken
                            break  # Do not iterate through other applications!

                        elif btm_available_capacity_current > 0 and btm_available_capacity_current < -temp_delta:
                            sub_system_states[btm_appl].soc = 0  # take everything from next application
                            temp_delta = temp_delta + btm_available_capacity_current  # reduce temp_delta before checking next BTM application

                    # in the end, all temp_delta should have been taken from other applications. There should be nothing left.
                    if temp_delta != 0 or subsystem_counter == indices_btm[-1]:
                        print("Warning: NOT ZERO! One SOC smaller than 0%")
                        # Take from prior application. Should not happen! But has to be corrected to keep system SOC correct!
                        # This might happen because of degradation! Fewer total_capacity is available.
                        for btm_appl in [x for x in indices_btm if x < subsystem_counter]:  # go through prior applications to borrow from them
                            btm_available_capacity_current = np.round(sub_system_states[btm_appl].soc *
                                                                      sub_system_states[btm_appl].capacity, 6)

                            if btm_available_capacity_current > 0 and btm_available_capacity_current >= -temp_delta:
                                # Reduce next BTM application SOC by desired capacity
                                sub_system_states[btm_appl].soc += temp_delta / sub_system_states[btm_appl].capacity
                                break  # Do not iterate through other applications!

                            elif btm_available_capacity_current > 0 and btm_available_capacity_current < -temp_delta:
                                sub_system_states[btm_appl].soc = 0  # take everything from next application
                                temp_delta = temp_delta + btm_available_capacity_current  # reduce temp_delta before checking next BTM application

                elif btm_ftm[subsystem_counter] == 1:  # FTM
                    for ftm_appl in [x for x in indices_ftm if x > subsystem_counter]:  # go through next applications to borrow from them
                        ftm_available_capacity_current = np.round(sub_system_states[ftm_appl].soc * \
                                                                  sub_system_states[ftm_appl].capacity, 6)

                        if ftm_available_capacity_current > 0 and ftm_available_capacity_current >= -temp_delta:
                            # Reduce next BTM application SOC by desired capacity
                            sub_system_states[ftm_appl].soc += temp_delta / sub_system_states[ftm_appl].capacity
                            temp_delta = 0  # because complete temp_delta is taken
                            break  # Do not iterate through other applications!

                        elif ftm_available_capacity_current > 0 and ftm_available_capacity_current < -temp_delta:
                            sub_system_states[ftm_appl].soc = 0  # take everything from next application
                            temp_delta = temp_delta + ftm_available_capacity_current  # reduce temp_delta before checking next BTM application

                    # in the end, all temp_delta should have been taken from other applications. There should be nothing left.
                    if temp_delta != 0 or subsystem_counter == indices_ftm[-1]:
                        print("Warning: NOT ZERO! One SOC smaller than 0%")
                        # Take from prior application. Should not happen! But has to be corrected to keep system SOC correct!
                        # This might happen because of degradation! Fewer total_capacity is available.
                        for ftm_appl in [x for x in indices_ftm if x < subsystem_counter]:  # go through prior applications to borrow from them
                            ftm_available_capacity_current = np.round(sub_system_states[ftm_appl].soc * \
                                                                      sub_system_states[ftm_appl].capacity, 6)

                            if ftm_available_capacity_current > 0 and ftm_available_capacity_current >= -temp_delta:
                                # Reduce next BTM application SOC by desired capacity
                                sub_system_states[ftm_appl].soc += temp_delta / sub_system_states[ftm_appl].capacity
                                break  # Do not iterate through other applications!

                            elif ftm_available_capacity_current > 0 and ftm_available_capacity_current < -temp_delta:
                                sub_system_states[ftm_appl].soc = 0  # take everything from next application
                                temp_delta = temp_delta + ftm_available_capacity_current  # reduce temp_delta before checking next BTM application

            elif sub_system_states[subsystem_counter].soc > 1:  # Energy was charged to other BTM pots
                temp_delta = (sub_system_states[subsystem_counter].soc - 1) * \
                             sub_system_states[subsystem_counter].capacity
                sub_system_states[subsystem_counter].soc = 1

                if btm_ftm[subsystem_counter] == 0:  # BTM
                    for btm_appl in [x for x in indices_btm if x > subsystem_counter]:  # go through next applications to borrow from them
                        btm_free_space_capacity_current = np.round((1 - sub_system_states[btm_appl].soc) * \
                                                                   sub_system_states[btm_appl].capacity, 6)

                        if btm_free_space_capacity_current > 0 and btm_free_space_capacity_current >= temp_delta:
                            # Increase next BTM application SOC by desired capacity
                            sub_system_states[btm_appl].soc += temp_delta / sub_system_states[btm_appl].capacity
                            temp_delta = 0
                            break  # Do not iterate through other applications!

                        elif btm_free_space_capacity_current > 0 and btm_free_space_capacity_current < temp_delta:
                            sub_system_states[btm_appl].soc = 1  # charge everything in next sub-application-storage
                            temp_delta = temp_delta - btm_free_space_capacity_current  # reduce temp_delta before checking next BTM application

                        # in the end, all temp_delta should have been taken from other applications. There should be nothing left.
                    if temp_delta != 0 or subsystem_counter == indices_btm[-1]:
                        print("Warning: NOT ZERO! One SOC smaller than 0%")
                        for btm_appl in [x for x in indices_btm if x < subsystem_counter]:  # go through next applications to borrow from them
                            btm_free_space_capacity_current = np.round((1 - sub_system_states[btm_appl].soc) * \
                                                                       sub_system_states[btm_appl].capacity, 6)


                            if btm_free_space_capacity_current > 0 and btm_free_space_capacity_current >= temp_delta:
                                # Increase next BTM application SOC by desired capacity
                                sub_system_states[btm_appl].soc += temp_delta / sub_system_states[btm_appl].capacity
                                break  # Do not iterate through other applications!

                            elif btm_free_space_capacity_current > 0 and btm_free_space_capacity_current < temp_delta:
                                sub_system_states[btm_appl].soc = 1  # charge everything in next sub-application-storage
                                temp_delta = temp_delta - btm_free_space_capacity_current  # reduce temp_delta before checking next BTM application

                elif btm_ftm[subsystem_counter] == 1:  # FTM
                    for ftm_appl in [x for x in indices_ftm if x > subsystem_counter]:  # go through next applications to borrow from them
                        ftm_free_space_capacity_current = np.round((1 - sub_system_states[ftm_appl].soc) * \
                                                                   sub_system_states[ftm_appl].capacity, 6)

                        if ftm_free_space_capacity_current > 0 and ftm_free_space_capacity_current >= temp_delta:
                            # Increase next BTM application SOC by desired capacity
                            sub_system_states[ftm_appl].soc += temp_delta / sub_system_states[ftm_appl].capacity
                            temp_delta = 0
                            break  # Do not iterate through other applications!

                        elif ftm_free_space_capacity_current > 0 and ftm_free_space_capacity_current < temp_delta:
                            sub_system_states[ftm_appl].soc = 1  # charge everything in next sub-application-storage
                            temp_delta = temp_delta - ftm_free_space_capacity_current  # reduce temp_delta before checking next BTM application

                        # in the end, all temp_delta should have been taken from other applications. There should be nothing left.
                    if temp_delta != 0 or subsystem_counter == indices_ftm[-1]:
                        print("Warning: NOT ZERO! One SOC smaller than 0%")
                        for ftm_appl in [x for x in indices_ftm if x < subsystem_counter]:  # go through next applications to borrow from them
                            ftm_free_space_capacity_current = np.round((1 - sub_system_states[ftm_appl].soc) * \
                                                                       sub_system_states[ftm_appl].capacity, 6)


                            if ftm_free_space_capacity_current > 0 and ftm_free_space_capacity_current >= temp_delta:
                                # Increase next BTM application SOC by desired capacity
                                sub_system_states[ftm_appl].soc += temp_delta / sub_system_states[ftm_appl].capacity
                                break  # Do not iterate through other applications!

                            elif ftm_free_space_capacity_current > 0 and ftm_free_space_capacity_current < temp_delta:
                                sub_system_states[ftm_appl].soc = 1  # charge everything in next sub-application-storage
                                temp_delta = temp_delta - ftm_free_space_capacity_current  # reduce temp_delta before checking next BTM application

        return sub_system_states

    def check_subsystem_soc_allows_power(self, indices_btm, indices_ftm, btm_ftm, power_requested, sub_system_states, borrowed_capacity):
        for subsystem_counter in range(len(power_requested)):
            current_capacity = sub_system_states[subsystem_counter].soc * \
                               sub_system_states[subsystem_counter].capacity - borrowed_capacity[subsystem_counter]  # Current available capacity, can't be zero
            desired_capacity_change = power_requested[subsystem_counter] * self.__ts / 3600
            desired_capacity = np.round(current_capacity + desired_capacity_change, 6)  # desired capacity if sub system is charged/discharged with requested power
            if btm_ftm[subsystem_counter] == 0:  # BTM
                if desired_capacity < 0:  # Discharge
                    # Available capacity of BTM applications that are in the ranking after current application:
                    for btm_appl in [x for x in indices_btm if x > subsystem_counter]:
                        btm_available_capacity_current = np.round(sub_system_states[btm_appl].soc * \
                                                                  sub_system_states[btm_appl].capacity, 6)

                        if btm_available_capacity_current > 0 and btm_available_capacity_current >= -desired_capacity:
                            # 1. Power is possible -> is already saved in self.__power_requested[subsystem_counter]
                            # 2. Reduce next BTM application by desired capacity
                            borrowed_capacity[btm_appl] -= desired_capacity
                            desired_capacity = 0
                            break  # Do not iterate through other applications!

                        elif btm_available_capacity_current > 0 and btm_available_capacity_current < -desired_capacity: # gesamte Energie zu gering
                            borrowed_capacity[btm_appl] = btm_available_capacity_current
                            desired_capacity = desired_capacity + btm_available_capacity_current
                    # After loop: Adjustment of power requested necessary because it can not
                    # borrow enough capacity from lower ranked applications
                    if desired_capacity < 0:
                        real_capacity_change = np.round(desired_capacity_change - desired_capacity, 6)
                        power_requested[subsystem_counter] = real_capacity_change / self.__ts * 3600

                elif desired_capacity > sub_system_states[subsystem_counter].capacity:  # Charge
                    # Free space capacity of BTM applications that are in the ranking after current application:
                    surplus_of_capacity = desired_capacity - sub_system_states[subsystem_counter].capacity
                    # Peak-Shaving was trying to charge rest in PVGreedy-application subsystem!
                    for btm_appl in [x for x in indices_btm if x > subsystem_counter]:
                        btm_free_space_capacity_current = np.round((1 - sub_system_states[btm_appl].soc) * \
                                                                    sub_system_states[btm_appl].capacity, 6)

                        if btm_free_space_capacity_current > 0 and btm_free_space_capacity_current >= surplus_of_capacity:
                        # 1. Power is possible -> was already saved in self.__power_requested[subsystem_counter]
                        # 2. Reduce next BTM application by desired capacity
                            borrowed_capacity[btm_appl] -= surplus_of_capacity
                            surplus_of_capacity = 0
                            break  # Do not iterate through other applications!

                        elif btm_free_space_capacity_current > 0 and btm_free_space_capacity_current < surplus_of_capacity:
                            borrowed_capacity[btm_appl] = -btm_free_space_capacity_current
                            surplus_of_capacity = surplus_of_capacity - btm_free_space_capacity_current

                    # After loop
                    if surplus_of_capacity > 0:
                        real_capacity_change = np.round(desired_capacity_change - surplus_of_capacity, 6)
                        power_requested[subsystem_counter] = real_capacity_change / self.__ts * 3600

            elif btm_ftm[subsystem_counter] == 1:  # FTM
                if desired_capacity < 0:  # Discharge
                    # Available capacity of FTM applications that are in the ranking after current application:
                    for ftm_appl in [x for x in indices_ftm if x > subsystem_counter]:

                        ftm_available_capacity_current = np.round(sub_system_states[ftm_appl].soc * \
                                                                  sub_system_states[ftm_appl].capacity, 6)

                        if ftm_available_capacity_current > 0 and ftm_available_capacity_current >= -desired_capacity:
                            # 1. Power is possible -> was already saved in self.__power_requested[subsystem_counter]
                            # 2. Reduce next BTM application by desired capacity
                            borrowed_capacity[ftm_appl] -= desired_capacity
                            desired_capacity = 0
                            break  # Do not iterate through other applications!

                        elif ftm_available_capacity_current > 0 and ftm_available_capacity_current < -desired_capacity:
                            borrowed_capacity[ftm_appl] = ftm_available_capacity_current
                            desired_capacity = desired_capacity + ftm_available_capacity_current

                    # After loop: Adjustment of power requested necessary because it can not
                    # borrow enough capacity from lower ranked applications
                    if desired_capacity < 0:
                        real_capacity_change = np.round(desired_capacity_change - desired_capacity, 6)
                        power_requested[subsystem_counter] = real_capacity_change / self.__ts * 3600

                elif desired_capacity > sub_system_states[subsystem_counter].capacity:  # Charge
                    # Free space capacity of BTM applications that are in the ranking after current application:
                    surplus_of_capacity = desired_capacity - sub_system_states[subsystem_counter].capacity
                    # Peak-Shaving was trying to charge rest in PVGreedy-application subsystem!
                    for ftm_appl in [x for x in indices_ftm if x > subsystem_counter]:

                        ftm_free_space_capacity_current = np.round((1 - sub_system_states[ftm_appl].soc) * \
                                                                   sub_system_states[ftm_appl].capacity, 6)

                        if ftm_free_space_capacity_current > 0 and ftm_free_space_capacity_current >= surplus_of_capacity:
                            # 1. Power is possible -> was already saved in self.__power_requested[subsystem_counter]
                            # 2. Reduce next BTM application by desired capacity
                            borrowed_capacity[ftm_appl] -= surplus_of_capacity
                            surplus_of_capacity = 0
                            break  # Do not iterate through other applications!

                        elif ftm_free_space_capacity_current > 0 and ftm_free_space_capacity_current < surplus_of_capacity:
                            borrowed_capacity[ftm_appl] = -ftm_free_space_capacity_current
                            surplus_of_capacity = surplus_of_capacity - ftm_free_space_capacity_current

                    # After loop
                    if surplus_of_capacity > 0:
                        real_capacity_change = np.round(desired_capacity_change - surplus_of_capacity, 6)
                        power_requested[subsystem_counter] = real_capacity_change / self.__ts * 3600

        return power_requested, borrowed_capacity

    def calculate_power_considering_ranking(self, power_requested_list, max_power_allocated, btm_ftm):
        power_left = max_power_allocated  # charge and discharge power together

        indices_btm = [i for i, x in enumerate(btm_ftm) if x == 0]
        btm_power_requested = [power_requested_list[i] for i in indices_btm]
        indices_ftm = [i for i, x in enumerate(btm_ftm) if x == 1]
        ftm_power_requested = [power_requested_list[i] for i in indices_ftm]

        btm_charge_sum = sum([x for x in btm_power_requested if x > 0])
        btm_discharge_sum = sum([x for x in btm_power_requested if x < 0])
        btm_balance = np.sign(sum(btm_power_requested))  # 1 if sum>0: more charging, -1 if sum<0: more discharging, 0 if sum=0

        if all(item >= 0 for item in btm_power_requested) or all(item < 0 for item in btm_power_requested):
            flag_btm = 0
        else:
            flag_btm = 1  # Change of sign in BTM applications

        ftm_charge_sum = sum([x for x in ftm_power_requested if x > 0])
        ftm_discharge_sum = sum([x for x in ftm_power_requested if x < 0])
        ftm_balance = np.sign(sum(ftm_power_requested))  # 1 if sum>0: more charging, -1 if sum<0: more discharging, 0 if sum=0

        if all(item >= 0 for item in ftm_power_requested) or all(item < 0 for item in ftm_power_requested):
            flag_ftm = 0
        else:
            flag_ftm = 1  # Change of sign in BTM applications

        # In Order of ranking
        power_allocated = []
        for counter_powers in range(len(power_requested_list)):  # Already in order of ranking
            power_requested = power_requested_list[counter_powers]

            # Ask FTM/BTM
            if btm_ftm[counter_powers] == 0:  # BTM application
                power_temp, power_left_temp, charge_temp, discharge_temp = \
                    self.allocate_power(power_requested, power_left, btm_balance, flag_btm, btm_charge_sum, btm_discharge_sum)
                btm_charge_sum = charge_temp
                btm_discharge_sum = discharge_temp
                power_left = power_left_temp
                power_allocated.append(power_temp)
            elif btm_ftm[counter_powers] == 1:   # FTM application
                power_temp, power_left_temp, charge_temp, discharge_temp = \
                    self.allocate_power(power_requested, power_left, ftm_balance, flag_ftm, ftm_charge_sum, ftm_discharge_sum)
                ftm_charge_sum = charge_temp
                ftm_discharge_sum = discharge_temp
                power_left = power_left_temp
                power_allocated.append(power_temp)

        return power_allocated

    def allocate_power(self, power_requested, power_left, balance, flag, charge_sum, discharge_sum):
        power_allocated = 0
        if power_requested > 0:  # Charge

            # 1: sum of btm/ftm discharge power > sum of btm/ftm charge power -> charge power always available
            if flag == 1 and balance <= 0:  # for balance == 0 and -1
                power_allocated = power_requested

            # 2: requested power is smaller than discharge_sum -> requested power can be allocated
            elif power_requested <= -discharge_sum and balance == 1:
                power_allocated = power_requested
                discharge_sum -= -power_allocated

            # 3: requested power is bigger than discharge_sum
            # requested power gets remaining discharge power and the power then still required comes from power left
            elif power_requested > -discharge_sum and balance == 1:
                temp = power_requested
                temp -= -discharge_sum
                if temp < power_left:
                    power_allocated = power_requested
                    power_left -= power_allocated - (-discharge_sum)
                else:  # temp >= power_left
                    power_allocated = power_left + (-discharge_sum)
                    power_left = 0
                discharge_sum = 0

        elif power_requested < 0:

            # 1: sum of btm/ftm discharge power > sum of btm/ftm charge power -> discharge power always available
            if flag == 1 and balance >= 0:  # for balance == 0 and 1
                power_allocated = power_requested

            # 2: requested power is smaller then btm_charge_sum
            # -> requested power can be allocated
            elif power_requested >= -charge_sum and balance == -1:
                power_allocated = power_requested
                charge_sum -= -power_allocated

            # 3: requested power is bigger than charge_sum. -> rest needs to be taken from power_left
            elif -power_requested > charge_sum and balance == -1:
                temp = power_requested
                temp += charge_sum
                if temp > - power_left:
                    power_allocated = power_requested
                    power_left = power_left + power_allocated + charge_sum
                else:  # temp >= power_left
                    power_allocated = -power_left - charge_sum
                    power_left = 0
                charge_sum = 0

        else:  # power_requested = 0
            power_allocated = 0  # power_allocated.append(0)

        return power_allocated, power_left, charge_sum, discharge_sum

    def update(self, energy_management_state: EnergyManagementState) -> None:
        for strategy in self.__strategies:
            strategy.update(energy_management_state)


from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.plotting.energy_plotting import EnergyPlotting


class SankeyDiagram(EnergyPlotting):
    """
    SankeyDiagram handles the backend data processing for the  sankey_diagram method of PlotlyPlotting.
    This class dynamically creates nodes, sources, targets, and values of flow quantities for the links as per the
    BESS configuration in SimSES.
    The class also creates all node labels and assigns colors to the links and nodes.
    """

    NODE_LABELS: str = 'NODE_LABELS'
    SOURCES: str = 'SOURCES'
    TARGETS: str = 'TARGETS'

    # Node colors
    NODE_COLORS: str = 'NODE_COLORS'
    LINK_COLORS: str = 'LINK_COLORS'

    def __init__(self, data: SystemData, title: str, path: str):
        super().__init__(data, path)
        self.__title = title
        self.__figures: list = list()

    def create_node_links(self) -> dict:
        """
        Creates node links (sources, targets, and values) from provided data and node labels
        :param node_labels: node labels as dict
        :return: link attributes (node labels, node colors, link colors, sources, targets, values) as dict
        """
        node_labels: dict = self.create_categories()
        data: SystemData = self.get_data()

        sources = list()
        targets = list()
        values = list()

        # Charging

        # Grid to Transformer
        sources.append(node_labels[self.COMPONENTS_CH][self.GRID])
        targets.append(node_labels[self.COMPONENTS_CH][self.TRANSFORMER])
        values.append(data.charge_energy)

        # Transformer to losses, Aux, and AC/DC Converter
        sources.append(node_labels[self.COMPONENTS_CH][self.TRANSFORMER])
        targets.append(node_labels[self.LOSSES_CH][self.TRANSFORMER])
        values.append(0)  # Currently transformer implementation is lossless

        sources.append(node_labels[self.COMPONENTS_CH][self.TRANSFORMER])
        targets.append(node_labels[self.COMPONENTS_CH][self.AUXILIARY])
        values.append(data.aux_energy_charging)  # TODO check if segregation is correct

        sources.append(node_labels[self.COMPONENTS_CH][self.TRANSFORMER])
        targets.append(node_labels[self.COMPONENTS_CH][self.ACDC_CONVERTER])
        values.append(data.ac_pe_charging_energy)  # TODO check if segregation is correct

        # AC/DC Converter to losses, DC/DC Converter
        sources.append(node_labels[self.COMPONENTS_CH][self.ACDC_CONVERTER])
        targets.append(node_labels[self.LOSSES_CH][self.ACDC_CONVERTER])
        values.append(
            data.ac_pe_charging_energy - data.dc_power_charging_energy)  # TODO check if value is correct

        sources.append(node_labels[self.COMPONENTS_CH][self.ACDC_CONVERTER])
        targets.append(node_labels[self.COMPONENTS_CH][self.DCDC_CONVERTER])
        values.append(data.dc_power_charging_energy)  # TODO check if value is correct

        # DC/DC Converter to losses, Storage Technology
        sources.append(node_labels[self.COMPONENTS_CH][self.DCDC_CONVERTER])
        targets.append(node_labels[self.LOSSES_CH][self.DCDC_CONVERTER])
        values.append(
            data.dc_power_charging_energy - data.dc_power_storage_charging_energy)  # TODO check if value is correct

        sources.append(node_labels[self.COMPONENTS_CH][self.DCDC_CONVERTER])
        targets.append(node_labels[self.COMPONENTS_CH][self.STORAGE_TECHNOLOGY])
        values.append(data.dc_power_storage_charging_energy)  # TODO check if value is correct

        # Accounting for initial energy content in Storage Technology
        sources.append(node_labels[self.COMPONENTS_CH][self.INITIAL_ENERGY_CONTENT])
        targets.append(node_labels[self.COMPONENTS_CH][self.STORAGE_TECHNOLOGY])
        values.append(data.initial_energy_content)

        # Storage Technology (charging) to Storage Technology losses
        sources.append(node_labels[self.COMPONENTS_CH][self.STORAGE_TECHNOLOGY])
        targets.append(node_labels[self.LOSSES_CH][self.STORAGE_TECHNOLOGY])
        values.append(data.storage_technology_loss_energy)  # TODO check if segregation is correct

        # Accounting for final energy content in Storage Technnology
        sources.append(node_labels[self.COMPONENTS_CH][self.STORAGE_TECHNOLOGY])
        targets.append(node_labels[self.COMPONENTS_CH][self.FINAL_ENERGY_CONTENT])
        values.append(data.final_energy_content)

        # Discharging
        # Storage Technology (charging) to DCDC Converter (Discharging)
        sources.append(node_labels[self.COMPONENTS_CH][self.STORAGE_TECHNOLOGY])
        targets.append(node_labels[self.COMPONENTS_DCH][self.DCDC_CONVERTER])
        values.append(-1 * data.dc_power_storage_discharging_energy)

        # DCDC Converter to losses, ACDC Converter
        sources.append(node_labels[self.COMPONENTS_DCH][self.DCDC_CONVERTER])
        targets.append(node_labels[self.LOSSES_DCH][self.DCDC_CONVERTER])
        values.append(abs(data.dc_power_discharging_energy - data.dc_power_storage_discharging_energy))

        sources.append(node_labels[self.COMPONENTS_DCH][self.DCDC_CONVERTER])
        targets.append(node_labels[self.COMPONENTS_DCH][self.ACDC_CONVERTER])
        values.append(-1 * data.dc_power_discharging_energy)

        # ACDC Converter to losses, Transformer, Aux
        sources.append(node_labels[self.COMPONENTS_DCH][self.ACDC_CONVERTER])
        targets.append(node_labels[self.LOSSES_DCH][self.ACDC_CONVERTER])
        values.append(data.total_pe_losses_energy - (
                data.ac_pe_charging_energy - data.dc_power_charging_energy))  # TODO obtain values

        sources.append(node_labels[self.COMPONENTS_DCH][self.ACDC_CONVERTER])
        targets.append(node_labels[self.COMPONENTS_DCH][self.TRANSFORMER])
        values.append(data.discharge_energy)

        sources.append(node_labels[self.COMPONENTS_DCH][self.ACDC_CONVERTER])
        targets.append(node_labels[self.COMPONENTS_DCH][self.AUXILIARY])
        values.append(data.total_aux_losses_energy - data.aux_energy_charging)  # TODO obtain values

        # Transformer to losses, Grid
        sources.append(node_labels[self.COMPONENTS_DCH][self.TRANSFORMER])
        targets.append(node_labels[self.LOSSES_DCH][self.TRANSFORMER])
        values.append(0)  # Lossless transformer

        sources.append(node_labels[self.COMPONENTS_DCH][self.TRANSFORMER])
        targets.append(node_labels[self.COMPONENTS_DCH][self.GRID])
        values.append(data.discharge_energy)

        # Convert alphabetical connections to index numbers of labels
        labels = list()
        node_colors = list()
        for v in node_labels.values():
            for key in v.keys():
                node_colors.append(self.COLORS_DICT[key])
            for label in v.values():
                labels.append(label)

        source_indices = list()
        for source in sources:
            source_indices.append(labels.index(source))

        target_indices = list()
        link_colors = list()
        for target in targets:
            if self.LOSSES in target:
                link_colors.append(self.COLORS_DICT[self.LOSSES])
            else:
                link_colors.append(self.COLORS_DICT[self.LINKS])
            target_indices.append(labels.index(target))

        links = dict()
        links[self.NODE_LABELS] = labels
        links[self.NODE_COLORS] = node_colors
        links[self.SOURCES] = source_indices
        links[self.TARGETS] = target_indices
        links[self.VALUES] = values
        links[self.LINK_COLORS] = link_colors
        return links



from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.log import Logger
from simses.commons.profile.technical.frequency import FrequencyProfile
from simses.commons.profile.technical.technical import TechnicalProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class FrequencyContainmentReserve(OperationStrategy):
    """
    Operation strategy for providing FCR. It was developed according to the German regulatory framework .
    The requested charging and discharging power is proportional to the frequency deviation.
    Below 49.8 Hz or above 50.2 Hz the output power is set to the prequalified power. Within the frequency dead band
    around 50 Hz with +/-10 mHz the output power is set to 0 W. The degree of freedom to exceed the output power by a
    factor of 1.2 is used aiming to bring the SOC back to a predefined SOC set point.
    """

    def __init__(self, config: GeneralSimulationConfig, config_ems: EnergyManagementConfig, profile_config: ProfileConfig):
        super().__init__(OperationPriority.HIGH)
        self.__log: Logger = Logger(type(self).__name__)
        self.__frequency: TechnicalProfile = FrequencyProfile(config, profile_config)
        self.__max_fcr_power = config_ems.max_fcr_power  # Watt
        self.__soc_set = config_ems.soc_set  # pu

        self.__target_frequency = 50.0  # Hz
        self.__max_frequency_deviation = 0.2  # Hz
        self.__frequency_dead_band = 0.01  # Hz
        self.__frequency_dead_time = 30  # s
        self.__power_fcr_last_step = 0  # W
        self.__frequency_max_limit = 51.0  # Hz
        self.__frequency_min_limit = 49.0  # Hz

        if config.timestep > 1:
            self.__log.warn('Timestep is > 1s. Thus, the results are distorted and are not valid. '
                             'Rethink your timestep')

        self.__max_frequency = self.__target_frequency + self.__max_frequency_deviation  # Hz
        self.__min_frequency = self.__target_frequency - self.__max_frequency_deviation  # Hz

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        frequency = self.__frequency.next(time)

        if frequency > self.__frequency_max_limit or frequency < self.__frequency_min_limit:
            frequency = self.__target_frequency

        # Determine FCR power with power to frequency static in pu
        power_fcr = (frequency - self.__target_frequency) / self.__max_frequency_deviation
        # Limit FCR power in pu [-1,1]
        power_fcr = max(min(power_fcr, 1), -1)

        ''' Degree of freedom 1: Slope (neglected) '''
        # frequencySlope = np.sign(
        #     power_fcr - self.__power_fcr_last_step) / self.__frequency_dead_time  # frequencySlope > 0: more power charged / discharged than in the step before
        #
        # self.__frequency_dead_time -= 1
        # if battery_system_state.get_soc() < self.__soc_set and power_fcr < 0 and frequencySlope > 0:  # if SOC is less than optimum, try to discharge as little as possible
        #     power_fcr = max(self.__power_fcr_last_step - frequencySlope, power_fcr)
        # elif battery_system_state.get_soc() < self.__soc_set and power_fcr > 0 and frequencySlope < 0:  # if SOC is less than optimum, try to charge as much as possible
        #     power_fcr = max(self.__power_fcr_last_step + frequencySlope, power_fcr)
        # elif battery_system_state.get_soc() > self.__soc_set and power_fcr > 0 and frequencySlope > 0:  # if SOC is higher than optimum, try to charge as little as possible
        #     power_fcr = min(self.__power_fcr_last_step + frequencySlope, power_fcr)
        # elif battery_system_state.get_soc() > self.__soc_set and power_fcr < 0 and frequencySlope < 0:  # if SOC is higher than optimum, try to discharge as much as possible
        #     power_fcr = min(self.__power_fcr_last_step - frequencySlope, power_fcr)
        # else:
        #     self.__frequency_dead_time = 30  # reset dead time
        # self.__power_fcr_last_step = power_fcr  # Save last value before using further degrees of freedom

        ''' Degree of freedom 2: Overfulfillment until 120% if soc > soc_set and charging vice versa '''
        if (system_state.soc > self.__soc_set and power_fcr < 0) or \
                (system_state.soc < self.__soc_set and power_fcr > 0):
            power_fcr = power_fcr * 1.2

        ''' Degree of freedom 3: Using frequency dead band around 50 Hz with +/-10 mHz '''
        # Use dead band if soc > soc_set and charging is necessary according to the frequency
        # Use dead band if soc < soc_set and discharging is necessary according to the frequency
        if abs(frequency - self.__target_frequency) <= self.__frequency_dead_band:
            if (system_state.soc >= self.__soc_set and power_fcr > 0) or \
                    (system_state.soc <= self.__soc_set and power_fcr < 0):
                power_fcr = 0

        return power_fcr * self.__max_fcr_power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.fcr_max_power = self.__max_fcr_power

    def clear(self) -> None:
        self.__power_fcr_last_step = 0.0

    def close(self) -> None:
        self.__frequency.close()



from configparser import ConfigParser

import pytest

from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.battery_management_system.management_system import BatteryManagementSystem
from simses.technology.lithium_ion.cell.generic import GenericCell
from simses.technology.lithium_ion.cell.type import CellType

THRESHOLD: float = 1e-10

voltage: float = 600  # V
capacity: float = 60e3  # Wh
soh: float = 1.0  # p.u.

start_soc: float = 0.5  # p.u.
start_temperature: float = 300  # K

max_current: float = 205  # A (for Generic Cell in current setup)

time_step: float = 60  # s


battery_config: ConfigParser = ConfigParser()
battery_config.add_section('BATTERY')
battery_config.set('BATTERY', 'START_SOC', str(start_soc))
battery_config.set('BATTERY', 'MIN_SOC', str(0.0))
battery_config.set('BATTERY', 'MAX_SOC', str(1.0))
battery_config.set('BATTERY', 'EOL', str(0.6))
battery_config.set('BATTERY', 'START_SOH', str(soh))


@pytest.fixture()
def uut() -> BatteryManagementSystem:
    cell_type: CellType = GenericCell(voltage, capacity, soh, BatteryConfig(battery_config))
    return BatteryManagementSystem(cell_type, BatteryConfig(battery_config))


def create_battery_state() -> LithiumIonState:
    cell_type: CellType = GenericCell(voltage, capacity, soh, BatteryConfig(battery_config))
    bs = LithiumIonState(0, 0)
    bs.time = 0
    bs.soc = start_soc
    bs.temperature = start_temperature
    bs.voltage = cell_type.get_open_circuit_voltage(bs)
    bs.voltage_open_circuit = cell_type.get_open_circuit_voltage(bs)
    bs.nominal_voltage = cell_type.get_nominal_voltage()
    bs.internal_resistance = cell_type.get_internal_resistance(bs)
    bs.capacity = cell_type.get_capacity(bs) * cell_type.get_nominal_voltage() * cell_type.get_soh_start()
    bs.soh = cell_type.get_soh_start()
    bs.soe = bs.soc * bs.capacity
    bs.fulfillment = 1.0
    bs.max_charge_power = cell_type.get_max_current(bs) * bs.voltage
    bs.max_discharge_power = cell_type.get_min_current(bs) * bs.voltage
    return bs


@pytest.mark.parametrize('temperature, current, result',
                         [
                            (0, 100, 0),
                            (300, 100, 100),
                            (400, 100, 0)
                         ]
                         )
def test_temperature(uut: BatteryManagementSystem, temperature: float, current: float, result: float):
    battery_state: LithiumIonState = create_battery_state()
    battery_state.temperature = temperature
    battery_state.current = current
    uut.update(time_step, battery_state, current * battery_state.voltage)
    uut.close()
    assert abs(battery_state.current - result) <= THRESHOLD


@pytest.mark.parametrize('current, result',
                         [
                            (-400, -max_current),
                            (-100, -100),
                            (0, 0),
                            (100, 100),
                            (400, max_current)
                         ]
                         )
def test_current(uut: BatteryManagementSystem, current: float, result: float):
    battery_state: LithiumIonState = create_battery_state()
    battery_state.current = current
    uut.update(time_step, battery_state, current * battery_state.voltage)
    uut.close()
    assert abs(battery_state.current - result) <= THRESHOLD


@pytest.mark.parametrize('soc, current, result',
                         [
                            (-1, -100, 0),
                            (1, 100, 0),
                            (0.5, 200, 200)
                         ]
                         )
def test_soc(uut: BatteryManagementSystem, soc: float, current: float, result: float):
    battery_state: LithiumIonState = create_battery_state()
    battery_state.soc = soc
    battery_state.soe = soc * battery_state.capacity
    battery_state.current = current
    uut.update(time_step, battery_state, current * battery_state.voltage)
    uut.close()
    assert abs(battery_state.current - result) <= THRESHOLD


@pytest.mark.parametrize('current, result',
                         [
                            (-400, max_current/400.0),
                            (-100, 1.0),
                            (0, 1.0),
                            (100, 1.0),
                            (400, max_current/400.0)
                         ]
                         )
def test_fulfillment(uut: BatteryManagementSystem, current: float, result: float):
    battery_state: LithiumIonState = create_battery_state()
    battery_state.current = current
    v1 = battery_state.voltage_open_circuit + battery_state.internal_resistance * current
    uut.update(time_step, battery_state, current * v1)
    v2 = battery_state.voltage # voltage with adjusted current
    uut.close()
    result *= v2 / v1 # compensate voltage change from current reduction
    assert abs(battery_state.fulfillment - result) <= THRESHOLD



import math

from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.commons.log import Logger
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel

class LGMJ1_NMCCalendarDegradationModel(CalendarDegradationModel):

    def __init__(self, cell_type: CellType):
        super().__init__(cell_type)
        self.__log: Logger = Logger(type(self).__name__)

        self.__capacity_loss = 0
        self.__initial_capacity = self._cell.get_nominal_capacity()
        self.__capacity_loss_cal = cell_type.get_calendar_capacity_loss_start()

        self.__A_QLOSS = -0.00315  # constant
        self.__B_QLOSS = 0.4467  # constant

        self.__resistance_increase = 0
        self.__rinc_calendar = 0

        self.__A_RINC = 0.00012736  # constant

    # Source: Khiem Trad
    # D2.3 – Report containing aging test profiles and test results
    # Everlasting, February 2020

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        temp = battery_state.temperature

        qloss: float = self.__capacity_loss_cal  # in pu
        virtual_time: float = (((1 - qloss) * 100 - 100) / (self.__A_QLOSS * temp + self.__B_QLOSS)) ** 2

        rel_capacity_status = lambda time: ((self.__A_QLOSS * temp + self.__B_QLOSS) * math.sqrt(time) + 100) / 100

        total_time = virtual_time + time_passed / 86400  # in days

        capacity_loss = (1 - rel_capacity_status(total_time)) - self.__capacity_loss_cal
        # Total calendrical capacity loss
        self.__capacity_loss_cal += capacity_loss
        # Delta calendrical capacity loss
        self.__capacity_loss = capacity_loss * self.__initial_capacity

    # Source: Zilberman I., Sturm J., Jossen A.
    # Reversible self-discharge and calendar aging of 18650 nickel-rich silicon-graphite lithium-ion cells
    # Journal of Power Sources, https://doi.org/10.1016/j.jpowsour.2019.03.109

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        rinc = self.__rinc_calendar
        time_passed = time - battery_state.time
        virtual_time: float = rinc / self.__A_RINC
        rel_resistance_increase = lambda time: (self.__A_RINC * time)
        total_time = virtual_time + time_passed / 86400  # in days
        resistance_increase = (rel_resistance_increase(total_time)) - self.__rinc_calendar
        self.__rinc_calendar += resistance_increase
        self.__resistance_increase = resistance_increase

    def get_degradation(self) -> float:
        return self.__capacity_loss

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        pass



# import os
# from configparser import ConfigParser
# from datetime import datetime, timezone
#
# import pytest
#
# from simses.commons.config.simulation.general import GeneralSimulationConfig
# from simses.commons.config.simulation.profile import ProfileConfig
# from simses.commons.config.simulation.energy_management import EnergyManagementConfig
# from simses.commons.profile.power.file import FilePowerProfile
# from simses.commons.profile.technical.binary import BinaryProfile
# from simses.commons.state.system import SystemState
# from simses.commons.utils.utilities import remove_file
# from simses.logic.energy_management.strategy.basic.electric_vehicle import \
#     ElectricVehicle
# from pathlib import Path
#
# DELIMITER: str = ','
# HEADER: str = '# Unit: W'
# system_state: SystemState = SystemState(0, 0)
# START: int = 0
# STEP: int = 1
# FILE_NAME_LOAD: str = 'mockup_power_load_profile.csv'
# FILE_NAME_BINARY: str = 'mockup_power_home_profile.csv'
# FILE_PATH_LOAD = os.path.join(Path(os.path.dirname(__file__)).parent.parent, 'data\profile\\power\\' + FILE_NAME_LOAD)
# FILE_PATH_BINARY = os.path.join(Path(os.path.dirname(__file__)).parent.parent, 'data\profile\\technical\\' + FILE_NAME_BINARY)
#
#
# def create_load_file():
#     with open(FILE_PATH_LOAD, mode='w') as file:
#         file.write(HEADER + '\n')
#         for counter in range(END):
#             file.write(str(counter) + DELIMITER + str(POWER_VALUES[counter]) + '\n')
#
#     power_load_profile = FilePowerProfile(config=create_general_config(), filename=FILE_PATH_LOAD,
#                                           delimiter=DELIMITER, scaling_factor=1)
#     return power_load_profile
#
#
# def create_binary_file():
#     with open(FILE_PATH_BINARY, mode='w') as file:
#         file.write(HEADER + '\n')
#         for counter in range(END):
#             file.write(str(counter) + DELIMITER + str(BINARY_VALUES[counter]) + '\n')
#
#     binary_profile = BinaryProfile(config=create_general_config(), profile_config=create_profile_config())
#     return binary_profile
#
#
# def create_general_config():
#     conf: ConfigParser = ConfigParser()
#     conf.add_section('GENERAL')
#     date = datetime.fromtimestamp(START, timezone.utc)
#     conf.set('GENERAL', 'START', date.strftime("%Y-%m-%d %H:%M:%S"))
#     date = datetime.fromtimestamp(END, timezone.utc)
#     conf.set('GENERAL', 'END', date.strftime("%Y-%m-%d %H:%M:%S"))
#     conf.set('GENERAL', 'TIME_STEP', str(STEP))
#     return GeneralSimulationConfig(config=conf)
#
#
# def create_profile_config():
#     conf: ConfigParser = ConfigParser()
#     conf.add_section('PROFILE')
#     conf.set('PROFILE', 'LOAD_PROFILE', FILE_NAME_LOAD)
#     conf.set('PROFILE', 'BINARY_PROFILE', FILE_NAME_BINARY)
#     return ProfileConfig(config=conf)
#
#
# def create_ems_config(ev_charging_strategy):
#     conf: ConfigParser = ConfigParser()
#     conf.add_section('ENERGY_MANAGEMENT')
#     conf.set('ENERGY_MANAGEMENT', 'STRATEGY', 'ElectricVehicle')
#     conf.set('ENERGY_MANAGEMENT', 'EV_CHARGING_STRATEGY', ev_charging_strategy)
#     conf.set('ENERGY_MANAGEMENT', 'Max_Power', str(MAX_AC_POWER))
#     return EnergyManagementConfig(config=conf)
#
#
# @pytest.fixture(scope='module')
# def uut0():
#     power_load_profile = create_load_file()
#     binary_profile = create_binary_file()
#     ems_config = create_ems_config(EV_CHARGING_STRATEGY[0])
#     uut0 = ElectricVehicle(power_load_profile, binary_profile, ems_config)
#
#     yield uut0
#
#
# @pytest.fixture(scope='module')
# def uut1():
#     power_load_profile = create_load_file()
#     binary_profile = create_binary_file()
#     ems_config = create_ems_config(EV_CHARGING_STRATEGY[1])
#     uut1 = ElectricVehicle(power_load_profile, binary_profile, ems_config)
#
#     yield uut1
#
#
# POWER_VALUES: list = [5, 6, 7, 8, 9, 10]
# END: int = len(POWER_VALUES)
# BINARY_VALUES: list = [2, 0, 1, 1, 1, 1]
# MAX_AC_POWER: int = 17
# EV_CHARGING_STRATEGY: list = ['Uncontrolled', 'Smart']
#
# @pytest.mark.parametrize('strategy, time, soc, result',
#              [
#                 (EV_CHARGING_STRATEGY[0], 0, 0, 'ValueError'),       # 1: Binary Profil different value than 0 or 1
#                 (EV_CHARGING_STRATEGY[0], 1, 0, -6),                 # 2: EV on road => use 5W
#                 (EV_CHARGING_STRATEGY[0], 2, 1, 0),                  # 3: EV parked but SOC=100% => 0W
#                 (EV_CHARGING_STRATEGY[0], 3, 0, 17),                 # 4: EV parked and SOC<100% => 7W says power, but charged with 17W
#                 (EV_CHARGING_STRATEGY[1], 4, 0, 8.5),                # 5:
#                 (EV_CHARGING_STRATEGY[1], 5, 0, 8.5)                 # 6:
#              ]
#                          )
# def test_next(strategy, time, soc, result, uut0, uut1):
#
#     system_state.soc = soc
#     if strategy == EV_CHARGING_STRATEGY[0]:
#         if time==0:
#             with pytest.raises(ValueError):
#                 uut0.next(time, system_state)
#         else:
#             assert abs(uut0.next(time, system_state) - result) <= 1e-10
#
#     elif strategy == EV_CHARGING_STRATEGY[1]:
#         assert abs(uut1.next(time, system_state) - result) <= 1e-10
#
#     if time== END-1:
#         uut0.close()
#         uut1.close()
#         remove_file(FILE_PATH_BINARY)
#         remove_file(FILE_PATH_LOAD)
#



from simses.commons.data.data_handler import DataHandler
from simses.commons.state.parameters import SystemParameters
from simses.commons.state.system import SystemState
from simses.commons.state.technology.storage import StorageTechnologyState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.power_electronics.dcdc_converter.abstract_dcdc_converter import DcDcConverter
from simses.technology.storage import StorageTechnology


class StorageSystemDC:
    """
    DC storage system class incorporates a DCDC converter and a storage technology. In this class the power is split
    into current voltage via the DCDC converter based on the current storage technology state.
    """

    def __init__(self, system_id: int, storage_id: int, data_export: DataHandler, dcdc_converter: DcDcConverter,
                 storage_technology: StorageTechnology):
        self.__data_export: DataHandler = data_export
        self.__system_id: int = system_id
        self.__storage_id: int = storage_id
        # TODO Thermal model for dcdc converter needed? (MM, AP)
        self.__dcdc_converter: DcDcConverter = dcdc_converter
        self.__storage_technology: StorageTechnology = storage_technology
        self.__system_state: SystemState = self.__get_current_state()

    def update(self, time: float, dc_power: float) -> None:
        """
        Function to update the states of a DC storage system

        Parameters
        ----------
        time : current simulation time
        dc_power : DC target power (from ac storage system))

        Returns
        -------

        """
        # TODO implement current calculation in dcdc converter (MM)
        # self.__dcdc_converter.from_battery(dc_power, voltage_storage)
        # self.__dcdc_converter.to_battery(power, voltage_storage)

        voltage = self.__storage_technology.state.voltage
        self.__dcdc_converter.calculate_dc_current(dc_power, voltage)
        current: float = self.__dcdc_converter.dc_current
        self.__storage_technology.distribute_and_run(time, current, voltage)
        self.__system_state: SystemState = self.__get_current_state()

    def wait(self):
        """
        In case of multihreading in storage technologies, the upper storage system needs to wait for results.
        """
        self.__storage_technology.wait()

    def get_auxiliaries(self) -> [Auxiliary]:
        auxiliaries: [Auxiliary] = list()
        auxiliaries.extend(self.__dcdc_converter.get_auxiliaries())
        auxiliaries.extend(self.__storage_technology.get_auxiliaries())
        return auxiliaries

    @property
    def volume(self) -> float:
        """
        Volume of dc system in m3

        Returns
        -------

        """
        volume: float = 0.0
        volume += self.__storage_technology.volume
        volume += self.__dcdc_converter.volume
        return volume

    def __get_current_state(self) -> SystemState:
        system_state: SystemState = SystemState(self.__system_id, self.__storage_id)
        storage_state: StorageTechnologyState = self.__storage_technology.state
        system_state.time = storage_state.time
        system_state.ac_fulfillment = storage_state.fulfillment
        system_state.soc = storage_state.soc
        system_state.capacity = storage_state.capacity
        system_state.soh = storage_state.soh
        system_state.temperature = storage_state.temperature
        system_state.max_charge_power = min(storage_state.max_charge_power, self.__dcdc_converter.max_power)
        system_state.max_discharge_power = min(storage_state.max_discharge_power, self.__dcdc_converter.max_power)
        system_state.storage_power_loss = storage_state.power_loss
        # reverse calculation
        system_state.dc_power_storage = storage_state.voltage * storage_state.current
        self.__dcdc_converter.reverse_calculate_dc_current(system_state.dc_power_storage, storage_state.voltage)
        system_state.dc_power_intermediate_circuit = self.__dcdc_converter.dc_power
        system_state.dc_power_loss = self.__dcdc_converter.dc_power_loss
        system_state.dc_circuit_voltage = self.__dcdc_converter.intermediate_circuit_voltage
        system_state.dc_current = self.__dcdc_converter.dc_current
        return system_state

    @property
    def state(self) -> SystemState:
        """
        Function to write dc states into SystemState

        Parameters
        -------

        Returns
        -------
        SystemState
           SystemState with all values from a dc system
        """
        return self.__system_state

    def get_system_parameters(self):
        parameters: dict = dict()
        parameters[SystemParameters.SYSTEM] = type(self).__name__
        parameters[SystemParameters.ID] = str(self.__system_id) + '.' + str(self.__storage_id)
        parameters[SystemParameters.STORAGE_TECHNOLOGY] = type(self.__storage_technology).__name__
        parameters[SystemParameters.DCDC_CONVERTER] = type(self.__dcdc_converter).__name__
        parameters.update(self.__storage_technology.get_system_parameters())
        return parameters

    def get_storage_technology(self) -> StorageTechnology:
        return self.__storage_technology

    def get_dc_dc_converter(self) -> DcDcConverter:
        return self.__dcdc_converter

    def close(self) -> None:
        """
        Closing all open resources in a DC storage system

        Parameters
        ----------

        Returns
        -------

        """
        self.__storage_technology.close()
        self.__dcdc_converter.close()



from simses.commons.state.system import SystemState
from simses.logic.power_distribution.power_distributor import PowerDistributor


class EqualPowerDistributor(PowerDistributor):

    """
    EqualPowerDistributor distributes the power equally to all system independent of their current state.
    """

    def __init__(self):
        super().__init__()
        self.__number: float = 1.0

    def set(self, time: float, states: [SystemState], power) -> None:
        self.__number = len(states)

    def get_power_for(self, power: float, state: SystemState) -> float:
        return power / self.__number



import math
import pandas as pd
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel
from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig


class SonyLFPCalendarDegradationModel(CalendarDegradationModel):

    def __init__(self, cell_type: CellType, battery_data_config: BatteryDataConfig, battery_config: BatteryConfig):
        super().__init__(cell_type)
        self.__initial_capacity = self._cell.get_nominal_capacity()

        # Source SONY_US26650FTC1_Product Specification and Naumann, Maik, et al.
        # "Analysis and modeling of calendar aging of a commercial LiFePO4/graphite cell."
        # Journal of Energy Storage 17 (2018): 153-169.
        # DOI: https://doi.org/10.1016/j.est.2018.01.019

        self.__capacity_loss = 0
        self.__capacity_loss_calendar_relative = cell_type.get_calendar_capacity_loss_start()
        self.__resistance_increase = 0

        self.__TEMP_REF = 298.15  # K
        self.__SOC_REF = 1  # pu
        self.__R = 8.3144598  # J/(K*mol)

        # # original parameters
        # self.__A_QLOSS = -2059.8  # K
        # self.__B_QLOSS = 9.2644  # constant
        # self.__K_REF_QLOSS = 1.2571 * 10 ** (-5)  # pu*s^(-0.5)
        # self.__C_QLOSS = 2.8575  # constant
        # self.__D_QLOSS = 0.60225  # constant
        # self.__EA_QLOSS = 17126  # J/mol
        #
        # self.__A_RINC = -8638.8  # K
        # self.__B_RINC = 29.992  # constant
        # self.__K_REF_RINC = 3.4194 * 10 ** (-10)  # pu*s^(-0.5)
        # self.__C_RINC = -3.3903  # constant
        # self.__D_RINC = 1.5604  # constant
        # self.__EA_RINC = 71827  # J/mol

        # read in parameters from CSV
        model_number = battery_config.degradation_model_number
        values_cap_model = pd.read_csv(battery_data_config.lfp_sony_degradation_capacity_file, skiprows=lambda x: x not in [0, model_number]).values.tolist()[0]
        values_res_model = pd.read_csv(battery_data_config.lfp_sony_degradation_resistance_file, skiprows=lambda x: x not in [0, model_number]).values.tolist()[0]
        # print("\nUsing the following cyc degradation model: \tNumber: " + str(model_number) + "\tIdentifier: " +
        #       values_cap_model[0])
        self.__K_REF_QLOSS = values_cap_model[1]  # pu*s^(-0.5)
        self.__EA_QLOSS = values_cap_model[2]  # J/mol
        self.__C_QLOSS = values_cap_model[3]  # constant
        self.__D_QLOSS = values_cap_model[4]  # constant

        self.__K_REF_RINC = values_res_model[1]  # pu*s^(-0.5)
        self.__EA_RINC = values_res_model[2]  # J/mol
        self.__C_RINC = values_res_model[3]  # constant
        self.__D_RINC = values_res_model[4]  # constant

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        time_passed = time - battery_state.time
        soc = battery_state.soc
        temp = battery_state.temperature

        # calculate stress factor dependent coefficients
        k_temp_qloss = self.__K_REF_QLOSS * math.exp(-self.__EA_QLOSS / self.__R * (1 / temp - 1 / self.__TEMP_REF))
        k_soc_qloss = self.__C_QLOSS * (soc - 0.5) ** 3 + self.__D_QLOSS

        # calculate capacity loss based on virtual time and past calendar degradation
        virtual_time = (self.__capacity_loss_calendar_relative / (k_temp_qloss * k_soc_qloss)) ** 2  # seconds
        capacity_loss = k_temp_qloss * k_soc_qloss * (virtual_time + time_passed) ** 0.5  # total cal. qloss in p.u.
        capacity_loss -= self.__capacity_loss_calendar_relative  # relative qloss in pu in current timestep

        self.__capacity_loss_calendar_relative += capacity_loss  # tracking calendar capacity loss in p.u.
        self.__capacity_loss = capacity_loss * self.__initial_capacity  # relative qloss in Ah in current timestep

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        soc = battery_state.soc
        temp = battery_state.temperature
        time_passed = time - battery_state.time

        k_temp_qloss = self.__K_REF_RINC * math.exp(-self.__EA_RINC / self.__R * (1 / temp - 1 / self.__TEMP_REF))
        k_soc_qloss = self.__C_RINC * (soc - 0.5) ** 2 + self.__D_RINC

        resistance_increase = k_temp_qloss * k_soc_qloss * time_passed
        self.__resistance_increase = resistance_increase  # pu

    def get_degradation(self) -> float:
        return self.__capacity_loss

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        pass



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.control.pressure.no_pressure_controller import NoPressureController
from simses.technology.hydrogen.electrolyzer.pressure.abstract_model import PressureModel


class NoPressureModel(PressureModel):

    def __init__(self):
        super().__init__()
        self.__pressure_cathode_el = 0  # barg
        self.__pressure_anode_el = 0  # barg
        self.__pressure_cathode_desire_el = 0  # barg
        self.__pressure_anode_desire_el = 0  # barg
        self.__hydrogen_outflow = 0  # mol/s
        self.__oxygen_outflow = 0  # mol/s
        self.pressure_controller = NoPressureController()

    def calculate(self, time, state: ElectrolyzerState) -> None:
        self.__pressure_anode_el = state.pressure_anode
        self.__pressure_cathode_el = state.pressure_cathode
        hydrogen_produced = state.hydrogen_production
        oxygen_produced = state.oxygen_production
        self.__hydrogen_outflow = self.pressure_controller.calculate_n_h2_out(self.__pressure_cathode_el, self.__pressure_cathode_desire_el, hydrogen_produced, 0)
        self.__oxygen_outflow = self.pressure_controller.calculate_n_o2_out(self.__pressure_anode_el, self.__pressure_anode_desire_el, oxygen_produced)

    def get_pressure_anode(self) -> float:
        return self.__pressure_anode_el

    def get_pressure_cathode(self) -> float:
        return self.__pressure_cathode_el

    def get_h2_outflow(self) -> float:
        return self.__hydrogen_outflow

    def get_o2_outflow(self) -> float:
        return self.__oxygen_outflow

    def get_h2o_outflow_cathode(self) -> float:
        return 0

    def get_h2o_outflow_anode(self) -> float:
        return 0



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class NoCyclicDegradationModel(CyclicDegradationModel):

    def __init__(self):
        super().__init__()

    def calculate_resistance_increase(self, time: float, state: ElectrolyzerState) -> None:
        pass

    def calculate_exchange_current_dens_decrerase(self, state: ElectrolyzerState):
        pass

    def get_resistance_increase(self) -> float:
        return 0

    def get_exchange_current_dens_decrease(self) -> float:
        return 0

    def reset(self, hydrogen_state: ElectrolyzerState) -> None:
        pass

    def close(self) -> None:
        pass


from abc import ABC, abstractmethod


class AcDcConverter(ABC):

    def __init__(self, max_power: float):
        self.__MAX_POWER: float = max_power  # W
        self.__STANDBY_POWER_THRESHOLD: float = max_power * 0.001  # W

    @abstractmethod
    def to_ac(self, power: float, voltage: float) -> float:
        """
        Calculated dc power for discharging process

        Parameters
        ----------
        power : float
            requested ac power in W
        voltage : float
            dc voltage of intermediate circuit in V

        Returns
        -------
        float
            dc power in W

        """
        pass

    @abstractmethod
    def to_dc(self, power: float, voltage: float) -> float:
        """
        Calculated dc power for charging process

        Parameters
        ----------
        power : float
            requested ac power in W
        voltage : float
            dc voltage of intermediate circuit in V

        Returns
        -------
        float
            dc power in W

        """
        pass

    @abstractmethod
    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        """
        recalculates ac power in W, if the BMS limits the DC_power

        Parameters
        ----------
        voltage :
            voltage of intermediate circuit in V
        dc_power :
            dc power of intermediate circuit in W

        Returns
        -------
        float
            ac power in W

        """
        pass

    @abstractmethod
    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        """
        recalculates ac power in W, if the BMS limits the DC_power

        Parameters
        ----------
        voltage :
            voltage of intermediate circuit in V
        dc_power :
            dc power of intermediate circuit in W

        Returns
        -------
        float
            ac power in W

        """
        pass

    @property
    @abstractmethod
    def volume(self) -> float:
        """
        Volume of acdc converter in m3

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def mass(self) -> float:
        """
        Mass of acdc converter in kg

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def surface_area(self) -> float:
        """
        Surface area of acdc converter in m2

        Returns
        -------

        """
        pass

    @property
    def max_power(self) -> float:
        """
        returns the maximum ac power of the acdc converter

        Parameters
        -------

        Returns
        -------
            float
                maximum power of the acdc converter (ac power)

        """
        return self.__MAX_POWER

    @property
    def standby_power_threshold(self) -> float:
        """
        returns the ac power threshold for standby

        Returns
        -------

        """
        return self.__STANDBY_POWER_THRESHOLD

    @classmethod
    @abstractmethod
    def create_instance(cls, max_power: float, power_electronics_config: None):
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing all open resources in acdc converter

        Parameters
        ----------

        Returns
        -------

        """
        pass



from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.commons.state.parameters_reader import SystemParametersReader


class EnergyPlotting:

    LINKS: str = 'LINKS'
    VALUES: str = 'VALUES'

    # Component types
    GRID: str = 'Grid'
    TRANSFORMER: str = 'Transformer'
    AUXILIARY: str = 'Auxiliary System(s)'
    ACDC_CONVERTER: str = 'ACDC Converter(s)'
    DCDC_CONVERTER: str = 'DCDC Converter(s)'
    STORAGE_TECHNOLOGY: str = 'Storage Technology(-ies)'
    INITIAL_ENERGY_CONTENT: str = 'Initial Energy Content'
    FINAL_ENERGY_CONTENT: str = 'Final Energy Content'

    # Loss categories
    LOSSES: str = ' Losses'

    # Energy flow
    CHARGING: str = ' (ch)'
    DISCHARGING: str = ' (dch)'
    COMPONENTS_CH: str = 'COMPONENTS_CH'
    COMPONENTS_DCH: str = 'COMPONENTS_DCH'
    LOSSES_CH: str = 'Charging Losses'
    LOSSES_DCH: str = 'Discharging Losses'

    COLORS_DICT: dict = {GRID: Plotting.Color.BROWN,
                         TRANSFORMER: Plotting.Color.YELLOW,
                         AUXILIARY: Plotting.Color.BLUE,
                         ACDC_CONVERTER: Plotting.Color.CYAN,
                         DCDC_CONVERTER: Plotting.Color.VIOLET,
                         STORAGE_TECHNOLOGY: Plotting.Color.BRIGHT_GREEN,
                         INITIAL_ENERGY_CONTENT: Plotting.Color.BRIGHT_BLUE,
                         FINAL_ENERGY_CONTENT: Plotting.Color.BRIGHT_BLUE,
                         LOSSES: Plotting.Color.RED,
                         LINKS: Plotting.Color.GREY}

    def __init__(self, data: SystemData, path: str):
        self.__data = data
        self.__data_id: str = data.id
        self.__system_parameters_reader: SystemParametersReader = SystemParametersReader(path)

    def create_categories(self):
        """
        Creates dynamic labels for all nodes/categories representing various components and loss categories in the BESS
        :return: Node/category labels for charge/discharge, and losses as dict
        """
        # Current implementation represents one StorageSystemAC
        category_labels = dict()
        labels_components = dict()
        labels_components[self.GRID] = self.GRID
        labels_components[
            self.TRANSFORMER] = self.TRANSFORMER  # Assuming that a transformer is outside StorageSystemAC
        ac_system_id = int(float(self.__data_id))
        ac_system_id_str = str(ac_system_id) + ' '
        if ac_system_id == 0:
            labels_components[self.ACDC_CONVERTER] = self.ACDC_CONVERTER
            labels_components[self.AUXILIARY] = self.AUXILIARY
            labels_components[self.DCDC_CONVERTER] = self.DCDC_CONVERTER
            labels_components[self.STORAGE_TECHNOLOGY] = self.STORAGE_TECHNOLOGY
        else:
            labels_components[
                self.ACDC_CONVERTER] = ac_system_id_str + self.__system_parameters_reader.get_acdc_converter_type_for(
                ac_system_id)
            auxiliaries = ''
            for aux in self.__system_parameters_reader.get_auxiliaries_for(ac_system_id):
                auxiliaries += aux + ' '
            labels_components[self.AUXILIARY] = ac_system_id_str + auxiliaries

            number_dc_systems: int = self.__system_parameters_reader.get_number_of_storage_systems_dc(ac_system_id)
            dc_idx = 1
            dcdc_converters = ''
            storage_technologies = ''
            while dc_idx <= number_dc_systems:
                dc_system_id_str = str(ac_system_id) + '.' + str(dc_idx) + ' '
                dcdc_converters += dc_system_id_str + self.__system_parameters_reader.get_dcdc_converter_type_for(ac_system_id, dc_idx) + ' '
                storage_technologies += dc_system_id_str + self.__system_parameters_reader.get_storage_technology_for(
                    ac_system_id, dc_idx) + ' '
                dc_idx += 1
            labels_components[self.DCDC_CONVERTER] = dcdc_converters
            labels_components[self.STORAGE_TECHNOLOGY] = storage_technologies

        category_labels[self.COMPONENTS_CH] = {k: v + self.CHARGING for k, v in labels_components.items()}
        category_labels[self.COMPONENTS_CH][self.INITIAL_ENERGY_CONTENT] = self.INITIAL_ENERGY_CONTENT
        category_labels[self.COMPONENTS_CH][self.FINAL_ENERGY_CONTENT] = self.FINAL_ENERGY_CONTENT

        category_labels[self.COMPONENTS_DCH] = {k: v + self.DISCHARGING for k, v in labels_components.items()}

        category_labels[self.LOSSES_CH] = {k: v + self.LOSSES + self.CHARGING for k, v in labels_components.items()}
        category_labels[self.LOSSES_DCH] = {k: v + self.LOSSES + self.DISCHARGING for k, v in
                                            labels_components.items()}
        return category_labels

    def get_data(self) -> SystemData:
        return self.__data

    def get_data_id(self) -> str:
        return self.__data_id

    def system_parameters_reader(self) -> SystemParametersReader:
        return self.__system_parameters_reader



from simses.commons.state.system import SystemState
from simses.logic.power_distribution.power_distributor import PowerDistributor


class SocBasedPowerDistributor(PowerDistributor):

    """
    The SocBasedPowerDistributor calculates the power share to the systems via an inverse distance weighting algorithm
    according to the system SOC. In charge case, the system with lower SOC receives a higher load - in discharge case
    vice versa.
    """

    __SHARE = 1e-12

    def __init__(self):
        super().__init__()
        self.__reverse_sum: float = 0.0
        self.__sum: float = 0.0

    def set(self, time: float, states: [SystemState], power) -> None:
        sum = 0.0
        reverse_sum = 0.0
        for state in states:
            sum += self.__get_value_from(state)
            reverse_sum += 1.0 / self.__get_value_from(state)
        self.__sum = sum
        self.__reverse_sum = reverse_sum

    def get_power_for(self, power: float, state: SystemState) -> float:
        if self.__is_charge(power):
            # inverse distance weighting for charging
            share: float = (1.0 / self.__get_value_from(state)) / self.__reverse_sum
        else:
            share: float = self.__get_value_from(state) / self.__sum
        return power * share

    def __get_value_from(self, state: SystemState) -> float:
        return max(state.soc, self.__SHARE)

    def __is_charge(self, power: float) -> bool:
        return power > 0



from simses.system.power_electronics.acdc_converter.fix_efficiency import FixEfficiencyAcDcConverter


class NoLossAcDcConverter(FixEfficiencyAcDcConverter):

    def __init__(self, max_power: float):
        super().__init__(max_power, efficiency=1.0)

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return NoLossAcDcConverter(max_power)



import math
import pandas as pd
import scipy.interpolate
from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class MolicelNMCCyclicDegradationModel(CyclicDegradationModel):
    # Values based on MA Ni Chuanqin (EES, TUM) and adapted from Yulong Zhao in order to have the same structure
    # as in the aging models of Maik Naumann (SonyLFP)
    __DOC_IDX = 0
    __LENGTH_DOC_ARRAY = 1001

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_data_config: BatteryDataConfig):
        super().__init__(cell_type, cycle_detector)
        self.__log: Logger = Logger(type(self).__name__)
        self.__capacity_loss = 0
        self.__capacity_loss_cyclic = cell_type.get_cyclic_capacity_loss_start()
        self.__rinc_cyclic = 0
        self.__resistance_increase = 0

        self.__A_QLOSS = 1.1587  # constant
        self.__B_QLOSS = 1.569  # constant
        self.__C_QLOSS = 0.5562  # constant

        self.__A_RINC = 0.5562  # constant

        capacity_cyc_file: str = battery_data_config.nmc_molicel_capacity_cyc_file
        capacity_cyc = pd.read_csv(capacity_cyc_file)  # -
        capacity_cyc_mat = capacity_cyc.iloc[:self.__LENGTH_DOC_ARRAY, 1]
        doc_arr = capacity_cyc.iloc[:, self.__DOC_IDX]
        self.__capacity_cyc_interp1d = scipy.interpolate.interp1d(doc_arr, capacity_cyc_mat, kind='linear')

        ri_cyc_file: str = battery_data_config.nmc_molicel_ri_cyc_file
        ri_cyc = pd.read_csv(ri_cyc_file)  # -
        ri_cyc_mat = ri_cyc.iloc[:(self.__LENGTH_DOC_ARRAY + 1), 1]
        doc_arr = ri_cyc.iloc[:, self.__DOC_IDX]
        self.__ri_cyc_interp1d = scipy.interpolate.interp1d(doc_arr, ri_cyc_mat, kind='linear')

        self.__cycle_detector: CycleDetector = cycle_detector

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        crate: float = self.__cycle_detector.get_crate() * 3600 # in 1 / s -> *3600 -> in 1/h
        doc: float = self.__cycle_detector.get_depth_of_cycle() # DOC

        qloss: float = self.__capacity_loss_cyclic  # only cyclic losses
        qcell: float = battery_state.capacity / battery_state.nominal_voltage/ self._cell.get_parallel_scale()
        # single cell capacity in Ah

        if crate <= -0.5:
            k_capacity_cyc = self.get_stressfkt_ca_cyc(doc) / self.__A_QLOSS * (self.__A_QLOSS) \
                             ** (math.log2(crate / (-0.5)))
        elif abs(crate) < 0.5:
            k_capacity_cyc = self.get_stressfkt_ca_cyc(doc)
        else:
            k_capacity_cyc = self.get_stressfkt_ca_cyc(doc) * self.__B_QLOSS ** (math.log2(crate / 0.5))

        if k_capacity_cyc == 0.0:
            virtual_ChargeThroughput: float = 0
        else:
            virtual_ChargeThroughput: float = (qloss / k_capacity_cyc)**(1 / self.__C_QLOSS)

        capacity_loss = max(0,
                            k_capacity_cyc * (virtual_ChargeThroughput + doc*qcell)**self.__C_QLOSS - qloss)

        self.__capacity_loss_cyclic += capacity_loss  # pu
        self.__capacity_loss = capacity_loss * self._cell.get_nominal_capacity()  # Ah

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        doc: float = self.__cycle_detector.get_depth_of_cycle() # DOC

        rinc_cyclic: float = self.__rinc_cyclic
        k_ri_cyclic = self.get_stressfkt_ri_cyc(doc)
        if k_ri_cyclic == 0.0:
            virtual_ChargeThroughput: float = 0
        else:
            virtual_ChargeThroughput: float = (rinc_cyclic / k_ri_cyclic)**(1 / self.__A_RINC)

        qcell: float = battery_state.capacity / battery_state.nominal_voltage / self._cell.get_parallel_scale()
        # single cell capacity in Ah

        resistance_increase = max(0, k_ri_cyclic * (virtual_ChargeThroughput + doc*qcell)\
                              **self.__A_RINC - self.__rinc_cyclic)


        self.__rinc_cyclic += resistance_increase
        self.__resistance_increase = resistance_increase  # pu

    def get_stressfkt_ca_cyc(self, doc: float) -> float:
        """
        get the stress factor for cyclic aging capacity loss

        Parameters
        ----------
        doc : depth of cycle

        Returns
        -------
        float : stress parameters of cyclic aging (capacity loss)

        """
        doc = self.__check_doc_range(doc)
        return float(self.__capacity_cyc_interp1d(doc))

    def get_stressfkt_ri_cyc(self, doc: float) -> float:
        """
        get the stress factor for cyclic aging resistance increase

        Parameters
        ----------
        doc : depth of cycle

        Returns
        -------
        float : stress parameters of cyclic aging (resistance increase)

        """
        doc = self.__check_doc_range(doc)
        return float(self.__ri_cyc_interp1d(doc))

    def __check_doc_range(self, doc: float) -> float:
        if doc < 0.0:
            self.__log.warn(str(doc) + ' is out of interpolation range')
            return 0.0
        elif doc > 1.0:
            self.__log.warn(str(doc) + ' is out of interpolation range')
            return 1.0
        else:
            return doc

    def get_degradation(self) -> float:
        capacity_loss = self.__capacity_loss
        self.__capacity_loss = 0    # Set value to 0, because cyclic losses are not calculated in each step
        return capacity_loss

    def get_resistance_increase(self) -> float:
        resistance_increase = self.__resistance_increase
        self.__resistance_increase = 0 # Set value to 0, because cyclic losses are not calculated in each step
        return resistance_increase

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__capacity_loss_cyclic = 0
        self.__rinc_cyclic = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        self.__log.close()



from simses.commons.state.technology.storage import StorageTechnologyState


class LithiumIonState(StorageTechnologyState):
    """
    Current physical state of the lithium_ion with the main electrical parameters.
    """

    SYSTEM_AC_ID: str = 'StorageSystemAC'
    SYSTEM_DC_ID: str = 'StorageSystemDC'
    SOC: str = 'SOC in p.u.'
    SOE: str = 'State of energy in Wh'
    SOH: str = 'State of health in p.u.'
    VOLTAGE: str = 'Voltage in V'
    VOLTAGE_OPEN_CIRCUIT: str = 'Open circuit voltage in V'
    NOMINAL_VOLTAGE: str = 'Nominal voltage in V'
    CURRENT: str = 'Current in A'
    TEMPERATURE: str = 'Temperature in K'
    CAPACITY: str = 'Capacity in Wh'
    INTERNAL_RESISTANCE: str = 'Internal resistance in Ohm'
    RESISTANCE_INCREASE: str = 'R increase in p.u.'
    POWER_LOSS: str = 'P_loss in W'
    FULFILLMENT: str = 'Bat_ful in p.u.'
    # VOLTAGE_INPUT: str = 'Voltage input in V'

    C_RATE: str = 'C_rate for Cyclic Degradation'
    DELTA_DOD: str = 'Delta_DOD for Cyclic Degradation'

    MAX_CHARGE_POWER: str = 'Maximum charging power in W'
    MAX_DISCHARGE_POWER: str = 'Maximum discharging power in W'

    CAPACITY_LOSS_CYCLIC: str = 'Cyclic Capacity Loss in Wh'
    CAPACITY_LOSS_CALENDRIC: str = 'Calendric Capacity Loss in Wh'
    CAPACITY_LOSS_OTHER: str = 'Other Capacity Loss in Wh'
    RESISTANCE_INCREASE_CYCLIC: str = 'Cyclic R Increase in p.u.'
    RESISTANCE_INCREASE_CALENDRIC: str = 'Calendric R Increase in p.u.'
    RESISTANCE_INCREASE_OTHER: str = 'Other R Increase in p.u.'


    def __init__(self, system_id: int, storage_id: int):
        super().__init__()
        self._initialize()
        self.set(self.SYSTEM_AC_ID, system_id)
        self.set(self.SYSTEM_DC_ID, storage_id)

    @property
    def soc(self) -> float:
        return self.get(self.SOC)

    @soc.setter
    def soc(self, value: float) -> None:
        self.set(self.SOC, value)

    @property
    def soe(self) -> float:
        return self.get(self.SOE)

    @soe.setter
    def soe(self, value: float) -> None:
        self.set(self.SOE, value)

    @property
    def soh(self) -> float:
        return self.get(self.SOH)

    @soh.setter
    def soh(self, value: float) -> None:
        self.set(self.SOH, value)

    @property
    def voltage(self) -> float:
        return self.get(self.VOLTAGE)

    @voltage.setter
    def voltage(self, value: float) -> None:
        self.set(self.VOLTAGE, value)

    @property
    def voltage_open_circuit(self) -> float:
        return self.get(self.VOLTAGE_OPEN_CIRCUIT)

    @voltage_open_circuit.setter
    def voltage_open_circuit(self, value: float) -> None:
        self.set(self.VOLTAGE_OPEN_CIRCUIT, value)

    @property
    def nominal_voltage(self) -> float:
        return self.get(self.NOMINAL_VOLTAGE)

    @nominal_voltage.setter
    def nominal_voltage(self, value: float) -> None:
        self.set(self.NOMINAL_VOLTAGE, value)

    @property
    def current(self) -> float:
        return self.get(self.CURRENT)

    @current.setter
    def current(self, value: float) -> None:
        self.set(self.CURRENT, value)

    @property
    def temperature(self) -> float:
        return self.get(self.TEMPERATURE)

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.set(self.TEMPERATURE, value)

    @property
    def capacity(self) -> float:
        return self.get(self.CAPACITY)

    @capacity.setter
    def capacity(self, value: float) -> None:
        self.set(self.CAPACITY, value)

    @property
    def internal_resistance(self) -> float:
        return self.get(self.INTERNAL_RESISTANCE)

    @internal_resistance.setter
    def internal_resistance(self, value: float) -> None:
        self.set(self.INTERNAL_RESISTANCE, value)

    @property
    def resistance_increase(self) -> float:
        return self.get(self.RESISTANCE_INCREASE)

    @resistance_increase.setter
    def resistance_increase(self, value: float) -> None:
        self.set(self.RESISTANCE_INCREASE, value)

    @property
    def power_loss(self) -> float:
        return self.get(self.POWER_LOSS)

    @power_loss.setter
    def power_loss(self, value: float) -> None:
        self.set(self.POWER_LOSS, value)

    @property
    def fulfillment(self) -> float:
        return self.get(self.FULFILLMENT)

    @fulfillment.setter
    def fulfillment(self, value: float) -> None:
        self.set(self.FULFILLMENT, value)

    # @property
    # def voltage_input(self) -> float:
    #     return self.get(self.VOLTAGE_INPUT)

    # @voltage_input.setter
    # def voltage_input(self, value: float) -> None:
    #     self.set(self.VOLTAGE_INPUT, value)

    @property
    def id(self) -> str:
        return 'LIION' + str(self.get(self.SYSTEM_AC_ID)) + str(self.get(self.SYSTEM_DC_ID))

    @property
    def is_charge(self) -> bool:
        return self.current > 0

    @classmethod
    def sum_parallel(cls, battery_states: []):
        """
        Classmethod to calculate the combined BatteryState over several batteries connected in parallel.

        Parameters
        ----------
        battery_states : List
            List of BatteryState which are connected in parallel.

        Returns
        -------
        battery_state:
            Combined BatteryState over the parallel BatteryStates.
        """
        # TODO check and improvement necessary (MM)
        battery_state = LithiumIonState(0, 0)
        for bs in battery_states:
            battery_state.add(bs)
        # get non average values
        resistance = cls.sum_reverse(cls.INTERNAL_RESISTANCE, battery_states)
        current = battery_state.current
        capacity = battery_state.capacity
        power_loss = battery_state.power_loss
        max_charge = battery_state.max_charge_power
        max_discharge = battery_state.max_discharge_power
        # calculate capacity weighted values
        soc = 0
        soh = 0
        for state in battery_states:
            capacity_ratio = state.capacity / capacity
            soc += state.soc * capacity_ratio
            soh += state.soh * capacity_ratio
        # average values
        size = len(battery_states)
        battery_state.divide_by(size)
        # set capacity weighted values
        battery_state.soc = soc
        battery_state.soh = soh
        # set non average values
        battery_state.internal_resistance = resistance
        battery_state.current = current
        battery_state.capacity = capacity
        battery_state.power_loss = power_loss
        battery_state.max_charge_power = max_charge
        battery_state.max_discharge_power = max_discharge
        return battery_state

    @classmethod
    def sum_serial(cls, battery_states: []):
        """
        Classmethod to calculate the combined battery_state over several batteries connected in serial.

        Parameters
        ----------
        battery_states : List
            List of serial connected battery_states.

        Returns
        -------
        battery_state:
            Combined battery_state over the serial connected battery_states.
        """
        # TODO check and improvement necessary (MM)
        battery_state = LithiumIonState(0, 0)
        for bs in battery_states:
            battery_state.add(bs)
        # get non average values
        resistance = battery_state.internal_resistance
        voltage = battery_state.voltage
        nominal_voltage = battery_state.nominal_voltage
        power_loss = battery_state.power_loss
        # voltage_input = battery_state.voltage_input
        # average values
        size = len(battery_states)
        battery_state.divide_by(size)
        # set non average values
        battery_state.internal_resistance = resistance
        battery_state.voltage = voltage
        battery_state.nominal_voltage = nominal_voltage
        battery_state.power_loss = power_loss
        # battery_state.voltage_input = voltage_input
        return battery_state

    @property
    def c_rate(self) -> float:
        return self.get(self.C_RATE)

    @c_rate.setter
    def c_rate(self, value: float) -> None:
        self.set(self.C_RATE, value)

    @property
    def delta_dod(self) -> float:
        return self.get(self.DELTA_DOD)

    @delta_dod.setter
    def delta_dod(self, value: float) -> None:
        self.set(self.DELTA_DOD, value)

    @property
    def capacity_loss_cyclic(self) -> float:
        return self.get(self.CAPACITY_LOSS_CYCLIC)

    @capacity_loss_cyclic.setter
    def capacity_loss_cyclic(self, value: float) -> None:
        self.set(self.CAPACITY_LOSS_CYCLIC, value)

    @property
    def capacity_loss_calendric(self) -> float:
        return self.get(self.CAPACITY_LOSS_CALENDRIC)

    @capacity_loss_calendric.setter
    def capacity_loss_calendric(self, value: float) -> None:
        self.set(self.CAPACITY_LOSS_CALENDRIC, value)

    @property
    def capacity_loss_other(self) -> float:
        return self.get(self.CAPACITY_LOSS_OTHER)

    @capacity_loss_other.setter
    def capacity_loss_other(self, value: float) -> None:
        self.set(self.CAPACITY_LOSS_OTHER, value)

    @property
    def resistance_increase_cyclic(self) -> float:
        return self.get(self.RESISTANCE_INCREASE_CYCLIC)

    @resistance_increase_cyclic.setter
    def resistance_increase_cyclic(self, value: float) -> None:
        self.set(self.RESISTANCE_INCREASE_CYCLIC, value)

    @property
    def resistance_increase_calendric(self) -> float:
        return self.get(self.RESISTANCE_INCREASE_CALENDRIC)

    @resistance_increase_calendric.setter
    def resistance_increase_calendric(self, value: float) -> None:
        self.set(self.RESISTANCE_INCREASE_CALENDRIC, value)

    @property
    def resistance_increase_other(self) -> float:
        return self.get(self.RESISTANCE_INCREASE_OTHER)

    @resistance_increase_other.setter
    def resistance_increase_other(self, value: float) -> None:
        self.set(self.RESISTANCE_INCREASE_OTHER, value)

    @property
    def max_charge_power(self) -> float:
        return self.get(self.MAX_CHARGE_POWER)

    @max_charge_power.setter
    def max_charge_power(self, value: float) -> None:
        self.set(self.MAX_CHARGE_POWER, value)

    @property
    def max_discharge_power(self) -> float:
        return self.get(self.MAX_DISCHARGE_POWER)

    @max_discharge_power.setter
    def max_discharge_power(self, value: float) -> None:
        self.set(self.MAX_DISCHARGE_POWER, value)



from simses.system.thermal.ambient.ambient_thermal_model import AmbientThermalModel


class ConstantAmbientTemperature(AmbientThermalModel):

    """
    ConstantAmbientTemperature provides a constant temperature over time.
    """

    def __init__(self, temperature: float = 25):
        """
        Constructor of ConstantAmbientTemperature

        Parameters
        ----------
        temperature :
            temperature in in °C, default: 25 °C
        """
        super().__init__()
        self.__temperature = 273.15 + temperature  # K

    def get_temperature(self, time) -> float:
        return self.__temperature

    def get_initial_temperature(self) -> float:
        return self.__temperature

    def create_instance(self) -> AmbientThermalModel:
        """
        reinstantiates the AmbientThermalModel
        :return: AmbientThermalModel
        """
        temperature = self.__temperature - 273.15  # in °C
        return ConstantAmbientTemperature(temperature)

    def close(self):
        pass



import plotly.graph_objects as go
from plotly.subplots import make_subplots
import re
from operator import itemgetter

class SystemParameterFigureGenerator():

    def __init__(self, system_parameters: str):

        self.__fig = go.Figure
        self.__system_parameters = system_parameters
        self.__system_parameters_dict: dict = dict()
        self.__max_number_DC_subsystems: int = int()
        self.__grid_height: int = 0
        self.__grid_width: int = 0
        # system parameters keys
        self.__subsystem = 'subsystems'
        self.__acdc_converter = "acdc_converter"
        self.__power_distribution = "power_distribution"
        self.__dcdc_converter = "dcdc_converter"
        self.cell_type = "cell_type"
        self.__technology = "technology"
        self.__id = "id"
        # Grid elements
        self.__spec: list = list()
        self.__AC_IDs: list = list()
        self.__DC_tech: list = list()
        self.__DCDC_converters: list = list()
        self.__ACDC_converters: list = list()
        self.__DC_power_distribution: list = list()
        self.__grid_pos_AC_IDs: list = list()
        self.__grid_pos_DC_power_distribution: list = list()
        self.__grid_pos_DC_tech: list = list()

    def generate(self):
        self.system_parameters_as_dict()
        self.grid_generator()
        self.annotation_generator()
        self.figure_generator()
        self.boxes_generator()
        self.boxes_filler()
        self.beautify_fig()
        return self.__fig

    def system_parameters_as_dict(self):
        with open(self.__system_parameters, 'r') as system_parameters:
            lines: [str] = system_parameters.readlines()
            __system_parameters_str = ""
            for line in lines:
                __system_parameters_str += line
            system_parameters_as_string = __system_parameters_str
            global system_parameters_as_dict
            system_parameters_as_dict = "global system_parameters_as_dict \nsystem_parameters_as_dict="
            exec(system_parameters_as_dict + system_parameters_as_string[22:])
            size_of_AC_systems = []
            for i in range(len(system_parameters_as_dict[self.__subsystem])):
                size_of_AC_systems.append(len(system_parameters_as_dict[self.__subsystem][i][self.__subsystem]))
            self.__max_number_DC_subsystems = max(size_of_AC_systems)
            self.__system_parameters_dict = system_parameters_as_dict

    def grid_generator(self):
        self.__grid_width = len(self.__system_parameters_dict[self.__subsystem]) * 3
        self.__grid_height = self.__max_number_DC_subsystems * 2 + 2
        AC_power_distribution = [{"colspan": self.__grid_width}]
        AC_power_distribution += [None] * (self.__grid_width - 1)
        AC_line = []
        DC_matrix = []
        for i in range(len(self.__system_parameters_dict[self.__subsystem])):
            AC_system = self.__system_parameters_dict[self.__subsystem][i]
            AC_line += [{"colspan": 3}, None, None]
            DC_power_distribution = [{"rowspan": self.__grid_height - 2}]
            DC_power_distribution += [None] * (self.__grid_height - 3)
            DC_columns = []
            for j in range(len(AC_system[self.__subsystem])):
                DC_columns += [{"colspan": 2, "rowspan": 2}, None]
            DC_col2 = [None] * len(AC_system[self.__subsystem] * 2)
            if len(DC_columns) < self.__grid_height - 2:
                dif = self.__grid_height - 2 - len(DC_columns)
                DC_columns += [None] * (dif)
                DC_col2 += [None] * (dif)
            DC_matrix.append(DC_power_distribution)
            DC_matrix.append(DC_columns)
            DC_matrix.append(DC_col2)
        DC_matrix = [[DC_matrix[j][i] for j in range(len(DC_matrix))] for i in range(len(DC_matrix[0]))]
        self.__spec.append(AC_power_distribution)
        self.__spec.append(AC_line)
        self.__spec.extend(DC_matrix)

    def annotation_generator(self):
        for i in range(len(self.__system_parameters_dict[self.__subsystem])):
            AC_system = self.__system_parameters_dict[self.__subsystem][i]
            self.__AC_IDs.append("AC_system " + AC_system['id'])
            self.__ACDC_converters.append(re.sub('([A-Z0-9])', r' \1', AC_system[self.__acdc_converter]))
            self.__DC_power_distribution.append(re.sub('([A-Z0-9])', r' \1', AC_system[self.__power_distribution]))
            for j in range(len(AC_system[self.__subsystem])):
                dc_system = AC_system[self.__subsystem][j]
                if self.cell_type in dc_system:
                    tech = re.sub('([A-Z0-9])', r'<br>\1', dc_system[self.__technology]) + " " + dc_system[
                        self.__id] + "<br>" + \
                           dc_system[self.cell_type]
                else:
                    tech = re.sub('([A-Z0-9])', r'<br>\1', dc_system[self.__technology]) + " " + dc_system[
                        self.__id]
                dcdc_converter = re.sub('([A-Z0-9])', r' \1', dc_system[self.__dcdc_converter])
                self.__DC_tech.append(tech)
                self.__DCDC_converters.append(dcdc_converter)

    def figure_generator(self):
        self.__fig = make_subplots(cols=self.__grid_width, rows=self.__grid_height, print_grid=False,
                                   specs=self.__spec, horizontal_spacing=0.02)

    def boxes_generator(self):
        for i in range(len(self.__spec)):
            for j in range(len(self.__spec[0])):
                grid = self.__spec[i][j]
                if grid != None:
                    self.__fig.add_trace(go.Scatter(x=[], y=[]), row=i + 1, col=j + 1)
                    if self.__spec[i][j] != None and grid["rowspan"] == (self.__grid_height - 2):
                        self.__grid_pos_DC_power_distribution.append((i + 1, j + 1))
                    if self.__spec[i][j] != None and grid["rowspan"] == 2 and grid["colspan"] == 2:
                        self.__grid_pos_DC_tech.append((i + 1, j + 1))
                    if self.__spec[i][j] != None and grid["colspan"] == 3:
                        self.__grid_pos_AC_IDs.append((i + 1, j + 1))
        self.__grid_pos_DC_tech.sort(key=itemgetter(1))
        # Handling the exception if there is only one AC system, or if the biggest AC system only has one DC subsystem
        if self.__grid_height == 4:
            self.__grid_pos_DC_power_distribution = self.__grid_pos_DC_power_distribution[::2]
        if self.__grid_width == 3:
            self.__grid_pos_AC_IDs = self.__grid_pos_AC_IDs[1::2]

    def boxes_filler(self):
        self.__fig.add_annotation(row=1, col=1, text=re.sub('([A-Z0-9])', r' \1', self.__system_parameters_dict[
            self.__power_distribution]), x=0.5,
                                  y=0.5, showarrow=False)
        for i in range(len(self.__grid_pos_AC_IDs)):
            self.__fig.add_annotation(row=self.__grid_pos_AC_IDs[i][0], col=self.__grid_pos_AC_IDs[i][1],
                                      text=self.__AC_IDs[i], x=0.5, y=0.5,
                                      showarrow=False)
        for i in range(len(self.__grid_pos_DC_power_distribution)):
            if self.__grid_height == 4:
                self.__fig.add_annotation(row=self.__grid_pos_DC_power_distribution[i][0],
                                          col=self.__grid_pos_DC_power_distribution[i][1],
                                          text=re.sub('([A-Z0-9])', r'<br>\1', self.__DC_power_distribution[i]),
                                          x=0.5, y=0.35, showarrow=False,
                                          font=dict(size=9))
                self.__fig.add_annotation(row=self.__grid_pos_DC_power_distribution[i][0],
                                          col=self.__grid_pos_DC_power_distribution[i][1],
                                          x=0.5, y=0.85, showarrow=False,
                                          text=re.sub('([A-Z0-9])', r'<br>\1',
                                                      self.__ACDC_converters[i].replace("Ac Dc", "ac-dc")),
                                          font=dict(size=8))
                self.__fig.add_shape(row=self.__grid_pos_DC_power_distribution[i][0],
                                     col=self.__grid_pos_DC_power_distribution[i][1],
                                     type="line", x0=0, y0=0.6, x1=1, y1=0.6,
                                     line=dict(color="White", width=3, dash="solid"))
            else:
                self.__fig.add_annotation(row=self.__grid_pos_DC_power_distribution[i][0],
                                          col=self.__grid_pos_DC_power_distribution[i][1],
                                          text=self.__DC_power_distribution[i], x=0.5, y=0.4, showarrow=False,
                                          textangle=270,
                                          font=dict(size=9 + self.__max_number_DC_subsystems))
                self.__fig.add_annotation(row=self.__grid_pos_DC_power_distribution[i][0],
                                          col=self.__grid_pos_DC_power_distribution[i][1],
                                          text=re.sub('([A-Z0-9])', r'<br>\1',
                                                      self.__ACDC_converters[i].replace("Ac Dc", "ac-dc")),
                                          x=0.5, y=0.9,
                                          showarrow=False, font=dict(size=8))
                self.__fig.add_shape(row=self.__grid_pos_DC_power_distribution[i][0],
                                     col=self.__grid_pos_DC_power_distribution[i][1],
                                     type="line", x0=0, y0=0.75, x1=1, y1=0.75,
                                     line=dict(color="White", width=3, dash="solid"))

        for i in range(len(self.__grid_pos_DC_tech)):
            self.__fig.add_annotation(row=self.__grid_pos_DC_tech[i][0], col=self.__grid_pos_DC_tech[i][1],
                                      text=self.__DC_tech[i], x=0.6, y=0.5,
                                      showarrow=False, font=dict(size=9))
            self.__fig.add_annotation(row=self.__grid_pos_DC_tech[i][0], col=self.__grid_pos_DC_tech[i][1],
                                      text=re.sub('([A-Z0-9])', r'<br>\1',
                                                  self.__DCDC_converters[i].replace("Dc Dc", "dc-dc")),
                                      x=0.1, y=0.5, showarrow=False, font=dict(size=9), textangle=270)
            self.__fig.add_shape(row=self.__grid_pos_DC_tech[i][0], col=self.__grid_pos_DC_tech[i][1], type="line",
                                 x0=0.25, y0=0, x1=0.25, y1=1, line=dict(color="White", width=3, dash="solid"))

    def beautify_fig(self):
        self.__fig.update_xaxes(range=[0, 1], showticklabels=False, showgrid=False, zeroline=False, fixedrange=True)
        self.__fig.update_yaxes(range=[0, 1], showticklabels=False, showgrid=False, zeroline=False, fixedrange=True)
        # self.__fig.update_layout(width=self.__grid_width * 90 + 350, height=self.__grid_height * 90,automargin=True)
        self.__fig.update_layout(autosize=True)
